import unittest
import os
from pathlib import Path
import shutil
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
from ds_capability import *
from ds_capability.components.commons import Commons
from ds_capability.intent.feature_engineer_intent import FeatureEngineerIntent
from ds_core.properties.property_manager import PropertyManager
from ds_capability.components.visualization import Visualisation as viz

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class FeatureEngineerModelTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_for_smoke(self):
        fe = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = fe.tools
        tbl = FeatureEngineer.from_memory().tools.get_synthetic_data_types(100)
        self.assertEqual(100, tbl.num_rows)
        
    def test_model_cat_cast(self):
        fe = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = fe.tools
        tbl = FeatureEngineer.from_memory().tools.get_synthetic_data_types(100)
        result = tools.model_cat_cast(tbl)
        schema = pa.schema([
            ("cat", pa.string()),
            ("num", pa.string()),
            ("int", pa.string()),
            ("bool", pa.string()),
            ("date", pa.string()),
            ("string", pa.string()),
        ])
        self.assertEqual(result.schema, schema)
        tbl = FeatureEngineer.from_memory().tools.get_synthetic_data_types(100)
        result = tools.model_cat_cast(tbl, cat_type=True)
        schema = pa.schema([
            ("cat", pa.dictionary(pa.int32(), pa.utf8())),
            ("num", pa.dictionary(pa.int32(), pa.utf8())),
            ("int", pa.dictionary(pa.int32(), pa.utf8())),
            ("bool", pa.dictionary(pa.int32(), pa.utf8())),
            ("date", pa.dictionary(pa.int32(), pa.utf8())),
            ("string", pa.dictionary(pa.int32(), pa.utf8())),
        ])
        self.assertEqual(result.schema, schema)

    def test_model_num_cast(self):
        fe = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = fe.tools
        tbl = FeatureEngineer.from_memory().tools.get_synthetic_data_types(100, category_encode=False)
        tbl = tools.model_cat_cast(tbl)
        result = tools.model_num_cast(tbl)
        schema = pa.schema([
            ("cat", pa.string()),
            ("num", pa.float64()),
            ("int", pa.int64()),
            ("bool", pa.int64()),
            ("date", pa.timestamp('ns')),
            ("string", pa.string()),
        ])
        self.assertEqual(result.schema, schema)
        tbl = pa.table([pa.array(['$1,000.00', '$56.34', '$0.45'])], names=['currency'])
        result = tools.model_num_cast(tbl, remove=['$', ','])
        self.assertEqual('double', result.column('currency').type)
        self.assertEqual([1000.0, 56.34, 0.45], result.column('currency').to_pylist())

    def test_model_sample_link(self):
        fe = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = fe.tools
        canonical = FeatureEngineer.from_memory().tools.get_synthetic_data_types(10, category_encode=False)
        other = FeatureEngineer.from_memory().tools.get_synthetic_data_types(5, category_encode=False)
        result = tools.model_concat_remote(canonical=canonical, other=other, headers=['int'], rename_map=['key'])
        self.assertCountEqual(result.column_names, canonical.column_names+['key'])
        result = tools.model_concat_remote(canonical=canonical, other=other, headers=['int', 'num'], rename_map={'int': 'key', 'num': 'prob'})
        self.assertCountEqual(result.column_names, canonical.column_names + ['key', 'prob'])
        result = tools.model_concat_remote(canonical=canonical, other=other, headers=['int'], rename_map=['key1'], multi_map={'key2': 'key1'})
        self.assertCountEqual(result.column_names, canonical.column_names + ['key1', 'key2'])
        self.assertTrue(result.column('key1').equals(result.column('key2')))


    def test_raise(self):
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))

    def get_party_table(self):
        document = [
            {"_id": "PI1832341", "interactionDate": {"startDateTime": "2023-01-02 04:49:06.955000", "endDateTime": "2023-01-02 04:50:35.130000"}, "status": "failed", "relatedParty": [{"_id": "C5089669", "role": "Customer", "engagedParty": {"_id": "I249908", "referredType": "Individual"}}, {"_id": "dclmappuser1", "role": "CSRAgent"}], "interactionItem": [{"item": {}}, {"item": {}}, {"item": {}}], "channel": [{"name": "DCLM", "role": "interaction creation", "referredType": "Channel"}], "productId": ["PR716796"], "type": "DeLinkProductUserRequest"},
            {"_id": "PI1832345", "interactionDate": {"startDateTime": "2023-01-02 04:52:47.834000", "endDateTime": "2023-01-02 04:52:53.122000"}, "status": "failed", "relatedParty": [{"_id": "C5089669", "role": "Customer", "engagedParty": {"_id": "I249908", "referredType": "Individual"}}, {"_id": "dclmappuser1", "role": "CSRAgent"}], "interactionItem": [{"item": {}}, {"item": {}}, {"item": {}}], "channel": [{"name": "DCLM", "role": "interaction creation", "referredType": "Channel"}], "productId": ["PR721074"], "type": "DeLinkProductUserRequest"},
            {"_id": "PI1832352", "interactionDate": {"startDateTime": "2023-01-02 04:58:06.792000", "endDateTime": "2023-01-02 04:59:42.650000"}, "status": "failed", "relatedParty": [{"_id": "dclmappuser1", "role": "CSRAgent"}, {"_id": "C5091676", "role": "Customer", "engagedParty": {"_id": "I253956", "referredType": "Individual"}}], "interactionItem": [{"item": {"requestedStartDate": "2023-01-02 04:59:38.298000"}}, {"item": {}}, {"item": {}}, {"item": {}}, {"item": {}}, {"item": {}}], "channel": [{"name": "DCLM", "role": "interaction creation", "referredType": "Channel"}], "productId": ["PR721185"], "type": "RegistrationRequest"},
            {"_id": "PI1832357", "interactionDate": {"startDateTime": "2023-01-02 05:06:12.877000", "endDateTime": "2023-01-02 05:07:49.597000"}, "status": "completed", "relatedParty": [{"_id": "dclmappuser1", "role": "CSRAgent"}, {"_id": "C5091679", "role": "Customer", "engagedParty": {"_id": "I253962", "referredType": "Individual"}}], "interactionItem": [{"item": {"requestedStartDate": "2023-01-02 05:07:44.329000"}}, {"item": {}}, {"item": {}}, {"item": {}}, {"item": {}}, {"item": {}}], "channel": [{"name": "DCLM", "role": "interaction creation", "referredType": "Channel"}], "productId": ["PR721189"], "type": "RegistrationRequest"},
            {"_id": "PI1832365", "interactionDate": {"startDateTime": "2023-01-02 05:13:52.453000", "endDateTime": "2023-01-02 05:15:12.898000"}, "status": "completed", "relatedParty": [{"_id": "dclmappuser1", "role": "CSRAgent"}, {"_id": "C5091686", "role": "Customer", "engagedParty": {"_id": "I253969", "referredType": "Individual"}}], "interactionItem": [{"item": {"requestedStartDate": "2023-01-02 05:15:08.082000"}}, {"item": {}}, {"item": {}}, {"item": {}}], "channel": [{"name": "DCLM", "role": "interaction creation", "referredType": "Channel"}], "productId": ["PR721193"], "type": "RegistrationRequest"},
            {"_id": "PI1832363", "interactionDate": {"startDateTime": "2023-01-02 05:13:50.672000", "endDateTime": "2023-01-02 05:16:20.100000"}, "status": "completed", "relatedParty": [{"_id": "dclmappuser1", "role": "CSRAgent"}, {"_id": "C5091684", "role": "Customer", "engagedParty": {"_id": "I253967", "referredType": "Individual"}}], "interactionItem": [{"item": {"requestedStartDate": "2023-01-02 05:16:14.449000"}}, {"item": {}}, {"item": {}}, {"item": {}}, {"item": {}}, {"item": {}}], "channel": [{"name": "DCLM", "role": "interaction creation", "referredType": "Channel"}], "productId": ["PR721199"], "type": "RegistrationRequest"},
            {"_id": "PI1832370", "interactionDate": {"startDateTime": "2023-01-02 05:16:10.685000", "endDateTime": "2023-01-02 05:17:44.885000"}, "status": "completed", "relatedParty": [{"_id": "dclmappuser1", "role": "CSRAgent"}, {"_id": "C5091690", "role": "Customer", "engagedParty": {"_id": "I253974", "referredType": "Individual"}}], "interactionItem": [{"item": {"requestedStartDate": "2023-01-02 05:17:39.901000"}}, {"item": {}}, {"item": {}}, {"item": {}}, {"item": {}}], "channel": [{"name": "DCLM", "role": "interaction creation", "referredType": "Channel"}], "productId": ["PR721203"], "type": "RegistrationRequest"},
            {"_id": "PI1832376", "interactionDate": {"startDateTime": "2023-01-02 05:28:06.660000", "endDateTime": "2023-01-02 05:28:14.543000"}, "status": "failed", "relatedParty": [{"_id": "C5089669", "role": "Customer", "engagedParty": {"_id": "I249908", "referredType": "Individual"}}, {"_id": "dclmappuser1", "role": "CSRAgent"}], "interactionItem": [{"item": {}}, {"item": {}}, {"item": {}}], "channel": [{"name": "DCLM", "role": "interaction creation", "referredType": "Channel"}], "productId": ["PR721074"], "type": "DeLinkProductUserRequest"},
            {"_id": "PI1832385", "interactionDate": {"startDateTime": "2023-01-02 05:35:44.427000"}, "status": "draft", "relatedParty": [{"_id": "C5091684", "role": "Customer", "engagedParty": {"_id": "I253967", "referredType": "Individual"}}, {"_id": "dclmappuser1", "role": "CSRAgent"}], "interactionItem": [{"item": {}}, {"item": {}}], "channel": [{"name": "DCLM", "role": "interaction creation", "referredType": "Channel"}], "productId": ["PR721199"], "type": "ModifyPlanProductRequest"},
            {"_id": "PI1832388", "interactionDate": {"startDateTime": "2023-01-02 05:35:59.615000"}, "status": "draft", "relatedParty": [{"_id": "C5091684", "role": "Customer", "engagedParty": {"_id": "I253967", "referredType": "Individual"}}, {"_id": "dclmappuser1", "role": "CSRAgent"}], "interactionItem": [{"item": {}}, {"item": {}}], "channel": [{"name": "DCLM", "role": "interaction creation", "referredType": "Channel"}], "productId": ["PR721199"], "type": "AddVasProductRequest"},
        ]
        return Commons.table_flatten(pa.Table.from_pylist(document))
    def get_indivdual_table(self):
        document = [
            {"_id": "I35138", "contactMedium": [{"medium": {"number": "50070028", "type": "mobile"}, "preferred": True}, {"medium": {"emailAddress": "mail@stc.com.kw", "type": "emailAddress"}, "preferred": True}], "gender": "M", "familyName": "Fouad", "givenName": "Fouad", "middleName": "Fouad"},
            {"_id": "I35145", "contactMedium": [{"medium": {"emailAddress": "panneer.rajadurai.c@solutions.com.kw", "type": "EmailAddress"}, "preferred": True}, {"medium": {"number": "51658317", "type": "mobile"}, "preferred": True}, {"medium": {"number": "51658317", "type": "whatsapp"}, "preferred": False}, {"medium": {"number": "51658317", "type": "telegram"}, "preferred": False}, {"medium": {"type": "telephone"}, "role": "AlternateNumber"}], "gender": "M", "familyName": "Jay", "givenName": "Bhuvana", "middleName": ""},
            {"_id": "I35146", "contactMedium": [{"medium": {"emailAddress": "bhuvana.stc21@gmail.com", "type": "EmailAddress"}, "preferred": True}, {"medium": {"type": "mobile"}, "preferred": False}, {"medium": {"type": "whatsapp"}, "preferred": False}, {"medium": {"type": "telegram"}, "preferred": False}], "gender": "F", "familyName": "CORP", "givenName": "TECNOTREE", "middleName": "LTD"},
            {"_id": "I35178", "contactMedium": [{"medium": {"emailAddress": "m.m.alkhoduri@outlook.com", "type": "emailAddress"}, "preferred": True}, {"medium": {"number": "55850055", "type": "mobile"}, "preferred": True}, {"medium": {"number": "55850055", "type": "whatsapp"}, "preferred": False}, {"medium": {"number": "55850055", "type": "telegram"}, "preferred": False}], "gender": "M", "familyName": "", "givenName": "MohammadalKoduri", "middleName": ""},
            {"_id": "I35179", "contactMedium": [{"medium": {"emailAddress": "ahb@bremenintl.com", "type": "emailAddress"}, "preferred": True}, {"medium": {"number": "51500014", "type": "mobile"}, "preferred": True}, {"medium": {"number": "51500014", "type": "whatsapp"}, "preferred": False}, {"medium": {"number": "51500014", "type": "telegram"}, "preferred": False}], "gender": "M", "familyName": "", "givenName": "AhmedBakhiet", "middleName": ""},
            {"_id": "I35180", "contactMedium": [{"medium": {"emailAddress": "test@gmail.com", "type": "emailAddress"}, "preferred": True}], "gender": "M", "familyName": "Admin", "givenName": "FakhrTest", "middleName": ""},
            {"_id": "I35181", "contactMedium": [], "gender": "M", "familyName": "test", "givenName": "test", "nationality": "", "middleName": ""}
        ]
        return Commons.table_flatten(pa.Table.from_pylist(document))

def tprint(t: pa.table, headers: [str, list]=None, d_type: [str, list]=None, regex: [str, list]=None):
    _ = Commons.filter_columns(t.slice(0,10), headers=headers, d_types=d_type, regex=regex)
    print(Commons.table_report(_).to_string())


if __name__ == '__main__':
    unittest.main()



import threading
from io import StringIO, BytesIO
import pyarrow as pa
import pyarrow.parquet as pq
import pyarrow.feather as feather
from pyarrow import csv
import os
from ds_core.handlers.abstract_handlers import AbstractSourceHandler, AbstractPersistHandler, HandlerFactory
from ds_core.handlers.abstract_handlers import ConnectorContract

__author__ = 'Darryl Oatridge'


class S3SourceHandler(AbstractSourceHandler):
    """ An Amazon AWS S3 source handler.

        URI Format:
            uri = 's3://<bucket>[/<path>]/<filename.ext>'

        Restrictions:
            - This does not use the AWS S3 Multipart Upload and is limited to 5GB files
    """

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Handler passing the connector_contract dictionary

        Extra Parameters in the ConnectorContract kwargs:
            - region_name (optional) session region name
            - profile_name (optional) session shared credentials file profile name
        """
        # required module import
        self.boto3 = HandlerFactory.get_module('boto3')
        self.botocore_exceptions = HandlerFactory.get_module('botocore.exceptions')
        super().__init__(connector_contract)
        cc_params = connector_contract.kwargs
        cc_params.update(connector_contract.query)  # Update kwargs with those in the uri query
        region_name = cc_params.pop('region_name', 'us-east-2')
        aws_access_key_id = cc_params.pop('aws_access_key_id', os.environ.get('AWS_ACCESS_KEY_ID'))
        aws_secret_access_key = cc_params.pop('aws_secret_access_key', os.environ.get('AWS_SECRET_ACCESS_KEY'))
        aws_session_token = cc_params.pop('aws_session_token', os.environ.get('AWS_SESSION_TOKEN'))
        profile_name = cc_params.pop('profile_name', None)
        self._session = self.boto3.Session(region_name=region_name, aws_access_key_id=aws_access_key_id,
                                      aws_secret_access_key=aws_secret_access_key, profile_name=profile_name,
                                      aws_session_token=aws_session_token)
        self._file_state = 0
        self._changed_flag = True
        self._lock = threading.Lock()

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['parquet', 'csv', 'tsv', 'txt', 'json', 'pickle']

    def exists(self) -> bool:
        """ Returns True is the file exists

        Extra Parameters in the ConnectorContract kwargs:
            - s3_list_params: (optional) a dictionary of additional s3 parameters directly passed to 'list_objects_v2'

        """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The S3 Source Connector Contract has not been set correctly")
        _cc = self.connector_contract
        if not isinstance(_cc, ConnectorContract):
            raise ValueError("The Python Source Connector Contract has not been set correctly")
        cc_params = _cc.kwargs
        cc_params.update(_cc.query)  # Update kwargs with those in the uri query
        s3_list_params = cc_params.pop('s3_list_params', {})
        if _cc.schema not in ['s3']:
            raise ValueError("The Connector Contract Schema has not been set correctly.")
        s3_client = self._session.client(_cc.schema)
        response = s3_client.list_objects_v2(Bucket=_cc.netloc, **s3_list_params)
        for obj in response.get('Contents', []):
            if obj['Key'] == _cc.path[1:]:
                return True
        return False

    def has_changed(self) -> bool:
        """ returns if the file has been modified

            - s3_get_params: (optional) a dictionary of additional s3 client parameters directly passed to 'get_object'
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The S3 Source Connector Contract has not been set correctly")
        _cc = self.connector_contract
        if not isinstance(_cc, ConnectorContract):
            raise ValueError("The Python Source Connector Contract has not been set correctly")
        cc_params = _cc.kwargs
        cc_params.update(_cc.query)  # Update kwargs with those in the uri query
        # pop all the extra params
        s3_get_params = cc_params.pop('s3_get_params', {})
        if _cc.schema not in ['s3']:
            raise ValueError("The Connector Contract Schema has not been set correctly.")
        s3_client = self._session.client(_cc.schema)
        try:
            s3_object = s3_client.get_object(Bucket=_cc.netloc, Key=_cc.path[1:], **s3_get_params)
        except self.botocore_exceptions.ClientError as e:
            code = e.response["Error"]["Code"]
            raise ConnectionError("Failed to retrieve the object from region '{}', bucket '{}' "
                                  "Key '{}' with error code '{}'".format(self._session.region_name, _cc.netloc,
                                                                         _cc.path[1:], code))
        state = s3_object.get('LastModified', 0)
        if state != self._file_state:
            self._changed_flag = True
            self._file_state = state
        return self._changed_flag

    def reset_changed(self, changed: bool = False):
        """ manual reset to say the file has been seen. This is automatically called if the file is loaded"""
        changed = changed if isinstance(changed, bool) else False
        self._changed_flag = changed

    def load_canonical(self, **kwargs) -> pa.Table:
        """Loads the canonical dataset, returning a Pandas DataFrame. This method utilises the pandas
        'pd.read_' methods and directly passes the kwargs to these methods.

        Extra Parameters in the ConnectorContract kwargs:
            - file_type: (optional) the type of the source file. if not set, inferred from the file extension
                            by default json files load as dict, to load as pandas use read_params '{as_dataframe: True}
            - encoding: (optional) the encoding of the s3 object body. Default 'utf-8'
            - s3_get_params: (optional) a dictionary of additional s3 client parameters directly passed to 'get_object'
            - read_params: (optional) value pair dict of parameters to pass to the read methods. Underlying
                           read methods the parameters are passed to are all pandas 'read_*', e.g. pd.read_csv

        """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The S3 Source Connector Contract has not been set correctly")
        _cc = self.connector_contract
        if not isinstance(_cc, ConnectorContract):
            raise ValueError("The Python Source Connector Contract has not been set correctly")
        _, _, _ext = _cc.address.rpartition('.')
        cc_params = _cc.kwargs
        cc_params.update(_cc.query)  # Update kwargs with those in the uri query
        cc_params.update(kwargs)     # Update with any passed though the call
        # pop all the extra params
        encoding = cc_params.pop('encoding', 'utf-8')
        file_type = cc_params.pop('file_type', _ext if len(_ext) > 0 else 'parquet')
        s3_get_params = cc_params.pop('s3_get_params', {})
        if file_type.lower() not in self.supported_types():
            raise ValueError("The file type {} is not recognised. "
                             "Set file_type parameter to a recognised source type".format(file_type))
        # session
        if _cc.schema not in ['s3']:
            raise ValueError("The Connector Contract Schema has not been set correctly.")
        s3_client = self._session.client(_cc.schema)
        try:
            s3_object = s3_client.get_object(Bucket=_cc.netloc, Key=_cc.path[1:], **s3_get_params)
        except self.botocore_exceptions.ClientError as e:
            code = e.response["Error"]["Code"]
            raise ConnectionError("Failed to retrieve the object from region '{}', bucket '{}' "
                                  "Key '{}' with error code '{}'".format(self._session.region_name, _cc.netloc,
                                                                         _cc.path[1:], code))
        resource_body = s3_object['Body'].read()
        with self._lock:
            if file_type.lower() in ['parquet', 'pq', 'pqt']:
                results = pq.read_table(BytesIO(resource_body), **cc_params)
            elif file_type.lower() in ['feather']:
                return feather.read_table(BytesIO(resource_body), **cc_params)
            elif file_type.lower() in ['csv', 'tsv', 'txt']:
                parse_options = csv.ParseOptions(**cc_params)
                results = csv.read_csv(BytesIO(resource_body), parse_options=parse_options)
            else:
                raise LookupError('The source format {} is not currently supported'.format(file_type))
        s3_client.close()
        return results


class S3PersistHandler(S3SourceHandler, AbstractPersistHandler):
    """ An Amazon AWS S3 source handler.

        URI Format:
            uri = 's3://<bucket>[/<path>]/<filename.ext>'

        Restrictions:
            - This does not use the AWS S3 Multipart Upload and is limited to 5GB files
    """

    def persist_canonical(self, canonical: pa.Table, **kwargs) -> bool:
        """ persists either the canonical dataset.

        Extra Parameters in the ConnectorContract kwargs:
            - file_type: (optional) the type of the source file. if not set, inferred from the file extension
            - s3_put_params: (optional) a dictionary of additional s3 client parameters directly passed to 'get_object'
            - write_params: (optional) value pair dict of parameters to pass to the write methods - pandas.to_csv,
                              pandas.to_json, pickle.dump and parquet.Table.from_pandas
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _uri = self.connector_contract.address
        return self.backup_canonical(uri=_uri, canonical=canonical, **kwargs)

    def backup_canonical(self, canonical: pa.Table, uri: str, **kwargs) -> bool:
        """ persists the canonical dataset as a backup to the specified URI resource. Note that only the
        address is taken from the URI and all other attributes are taken from the ConnectorContract

        Extra Parameters in the ConnectorContract kwargs:
            - file_type: (optional) the type of the source file. if not set, inferred from the file extension
            - s3_put_params: (optional) value pair dict of parameters to pass to the Boto3 put_object method
            - write_params: (optional) value pair dict of parameters to pass to the write methods - pandas.to_csv,
                              pandas.to_json, pickle.dump and parquet.Table.from_pandas
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The S3 Source Connector Contract has not been set correctly")
        _cc = self.connector_contract
        if not isinstance(_cc, ConnectorContract):
            raise ValueError("The Python Source Connector Contract has not been set correctly")
        schema, bucket, path = _cc.parse_address_elements(uri=uri)
        _, _, _ext = path.rpartition('.')
        cc_params = kwargs if isinstance(kwargs, dict) else _cc.kwargs
        cc_params.update(_cc.parse_query(uri=uri))
        # pop all the extra params
        s3_put_params = cc_params.pop('s3_put_params', _cc.kwargs.get('put_object_kw', {}))
        write_params = cc_params.pop('write_params', _cc.kwargs.get('write_kw', {}))
        file_type = cc_params.pop('file_type', _cc.kwargs.get('file_type', _ext if len(_ext) > 0 else 'pkl'))
        if _cc.schema not in ['s3']:
            raise ValueError("The Connector Contract Schema has not been set correctly.")
        s3_client = self._session.client(_cc.schema)
        # csv
        if file_type.lower() in ['csv', 'tsv', 'txt']:
            byte_obj = BytesIO()
            with self._lock:
                csv.write_csv(canonical, byte_obj)
                s3_client.put_object(Bucket=bucket, Key=path[1:], Body=byte_obj.getvalue(), **s3_put_params)
        # parquet
        elif file_type.lower() in ['parquet', 'pq', 'pqt']:
            byte_obj = BytesIO()
            with self._lock:
                pq.write_table(canonical, byte_obj)
                s3_client.put_object(Bucket=bucket, Key=path[1:], Body=byte_obj.getvalue(), **s3_put_params)
        elif file_type.lower() in ['feather']:
            byte_obj = BytesIO()
            with self._lock:
                feather.write_feather(canonical, byte_obj)
                s3_client.put_object(Bucket=bucket, Key=path[1:], Body=byte_obj.getvalue(), **s3_put_params)
                return True
        else:
            raise LookupError('The source format {} is not currently supported for write'.format(file_type))
        s3_client.close()
        return True

    def remove_canonical(self) -> bool:
        """ removes the URI named resource

        Extra Parameters in the ConnectorContract kwargs:
            - s3_del_params: (optional) value pair dict of parameters to pass to the Boto3 delete_object method
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The S3 Source Connector Contract has not been set correctly")
        _cc = self.connector_contract
        if not isinstance(_cc, ConnectorContract):
            raise ValueError("The Python Source Connector Contract has not been set correctly")
        cc_params = _cc.kwargs
        cc_params.update(_cc.query)  # Update kwargs with those in the uri query
        # pop all the extra params
        s3_del_params = cc_params.pop('s3_put_params', _cc.kwargs.get('put_object_kw', {}))
        if _cc.schema not in ['s3']:
            raise ValueError("The Connector Contract Schema has not been set correctly.")
        s3_client = self._session.client(_cc.schema)
        response = s3_client.response = s3_client.delete_object(Bucket=_cc.netloc, Key=_cc.path[1:], **s3_del_params)
        if response.get('RequestCharged') is None:
            return False
        return True



import unittest
import os
from pathlib import Path
import shutil
from pprint import pprint
import pyarrow as pa
import pyarrow.compute as pc
import pandas as pd
from ds_capability import FeatureBuild, Controller
from ds_capability.intent.feature_build_intent import FeatureBuildIntent
from ds_core.properties.property_manager import PropertyManager

from ds_capability.components.discovery import DataDiscovery


class ControllerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass
    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]

        os.environ['HADRON_PM_PATH'] = os.path.join('work', 'config')
        os.environ['HADRON_DEFAULT_PATH'] = os.path.join('work', 'data')
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except:
            pass
        PropertyManager._remove_all()
        tr = Transition.from_env('task1', has_contract=False)
        tr.set_source_uri("https://raw.githubusercontent.com/mwaskom/seaborn-data/master/titanic.csv")
        tr.set_persist()
        wr = Wrangle.from_env('task2' , has_contract=False)
        wr.set_source_uri(tr.get_persist_contract().raw_uri)
        wr.set_persist()
        controller = Controller.from_env(has_contract=False)
        controller.intent_model.transition(canonical=pd.DataFrame(), task_name='task1', intent_level='task1_tr')
        controller.intent_model.wrangle(canonical=pd.DataFrame(), task_name='task2', intent_level='task2_wr')

    def tearDown(self):
        try:
            shutil.rmtree('work')
        except:
            pass

    def test_smoke(self):
        """Basic smoke test"""
        Controller.from_env(has_contract=False)

    def test_run_controller(self):
        controller = Controller.from_env()
        # test errors
        with self.assertRaises(ValueError) as context:
            controller.run_controller(run_book='noname')
        self.assertTrue("The run book or intent level" in str(context.exception))
        controller.run_controller()
        self.assertEqual(['hadron_wrangle_task2_primary_persist.pickle'], os.listdir('work/data/'))

    def test_repeat_iterations(self):
        wr = Wrangle.from_env('task2')
        wr.set_persist(wr.pm.file_pattern(name='tester', prefix='result1_', file_type='parquet'))
        controller = Controller.from_env()
        controller.run_controller(repeat=1)
        self.assertEqual(['result1_tester.parquet'], os.listdir('work/data/'))
        shutil.rmtree('work/data')
        os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        wr.set_persist(wr.pm.file_pattern(name='tester', prefix='result1_', file_type='parquet', stamped='ns'))
        controller.run_controller(repeat=3)
        self.assertEqual(3, len(os.listdir('work/data/')))
        shutil.rmtree('work/data')
        os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        controller.run_controller(repeat=2, sleep=1)
        self.assertEqual(2, len(os.listdir('work/data/')))
        shutil.rmtree('work/data')
        os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        controller.run_controller(repeat=2, sleep=1, run_time=4)
        self.assertEqual(4, len(os.listdir('work/data/')))

    def test_synthetic_with_no_source(self):
        shutil.rmtree('work/config')
        os.makedirs(os.environ['HADRON_PM_PATH'])
        PropertyManager._remove_all()
        builder = SyntheticBuilder.from_env('task3', has_contract=False)
        tools: SyntheticIntentModel = builder.tools
        builder.set_persist()
        df = pd.DataFrame(index=range(10))
        tools.model_noise(df, num_columns=5, column_name='noise')
        controller = Controller.from_env(has_contract=False)
        controller.intent_model.synthetic_builder(df, task_name='task3')
        controller.run_controller()
        self.assertIn(builder.CONNECTOR_PERSIST, builder.report_connectors(stylise=False)['connector_name'].to_list())
        self.assertNotIn(builder.CONNECTOR_SOURCE, builder.report_connectors(stylise=False)['connector_name'].to_list())

    def test_controller_log(self):
        controller = Controller.from_env()
        controller.run_controller(run_cycle_report='report.csv')
        df = controller.load_canonical(connector_name='run_cycle_report')
        control = ['start run-cycle 0', 'start task cycle 0', 'running task1_tr', 'canonical shape is (891, 15)',
                   'running task2_wr', 'canonical shape is (891, 15)', 'tasks complete', 'end of report']
        self.assertEqual(control, df['text'].to_list())
        controller.run_controller(run_time=3, run_cycle_report='report.csv')
        df = controller.load_canonical(connector_name='run_cycle_report')
        self.assertEqual(3, df.where(df['text'].str.startswith('start run-cycle')).dropna().shape[0])

    def test_controller_check_changed(self):
        tr = Transition.from_env('task1')
        tr.set_source(uri_file='sample.csv')
        tr.set_persist(uri_file='sample.csv')
        df = pd.DataFrame({'A': [1,2,3,4]})
        tr.save_persist_canonical(df)
        controller = Controller.from_env()
        controller.run_controller(repeat=2, source_check_uri=tr.get_persist_contract().raw_uri, run_cycle_report='report.csv')
        df = controller.load_canonical(connector_name='run_cycle_report')
        self.assertEqual(1, df.where(df['text'].str.startswith('start run-cycle')).dropna().shape[0])
        self.assertEqual(1, df.where(df['text'].str.startswith('Source has not changed')).dropna().shape[0])

    def test_controller_task_run(self):
        os.environ['HADRON_SYNTHETIC_DATA_SIZE'] = '2000'
        sb = SyntheticBuilder.from_env('data_types', has_contract=False)
        sb.set_persist()
        df = sb.tools.model_synthetic_data_types(canonical=1_000, extended=True, column_name='data_types')
        sb.run_component_pipeline(1_000)
        # controller
        controller = Controller.from_env(has_contract=False)
        controller.intent_model.synthetic_builder(canonical=0, task_name='data_types', intent_level='synthetic_data_types')
        run_book = [
            controller.runbook2dict(task='synthetic_data_types', source='${HADRON_SYNTHETIC_DATA_SIZE}'),
        ]
        controller.add_run_book(run_levels=run_book)
        controller.run_controller()
        result = sb.load_persist_canonical()
        self.assertEqual((2000, 27), result.shape)

    def test_capability_use_case(self):
        os.environ['TESTER_SOURCE_FILE'] = 'source/file.csv'

        tr = Transition.from_env('tester', has_contract=False)
        pm: TransitionPropertyManager = tr.pm
        tr.set_source_uri('${TESTER_SOURCE_FILE}')
        tr.set_persist()
        cc = pm.get_connector_contract(tr.CONNECTOR_SOURCE)
        print(cc.raw_handler)

        os.environ['TESTER_SOURCE_FILE'] = 'mysql://user:pass@localhost:3306/mydb/'
        cc = pm.get_connector_contract(tr.CONNECTOR_SOURCE)
        print(cc.raw_handler)


    def test_controller_use_case(self):
        os.environ['HADRON_SYNTHETIC_DATA_SIZE'] = '1000'
        os.environ['HADRON_SYNTHETIC_OUTCOME_URI'] = 'work/data/output.pq'

        controller = Controller.from_env(uri_pm_repo='https://raw.githubusercontent.com/project-hadron/hadron-asset-bank/master/contracts/commons/synthetic_data_types')

        controller.run_controller(run_cycle_report='hadron_controller_cycle_report.csv')
        report = controller.load_canonical(connector_name='run_cycle_report')
        print(report)

    def test_raise(self):
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))


if __name__ == '__main__':
    unittest.main()



from os.path import abspath, join, dirname
import time
from pathlib import Path
import pandas as pd
import numpy as np
import pyarrow as pa
from abc import ABC, abstractmethod
from ds_core.handlers.abstract_handlers import HandlerFactory

__author__ = 'Darryl Oatridge'


class AbstractSample(ABC):

    @staticmethod
    def _full_path(filename):
        return abspath(join(dirname(__file__), filename))

    @abstractmethod
    def __dir__(self):
        pass

    @staticmethod
    def _get_constant(reference: str, size: int=None, shuffle: bool=True, seed: int=None) -> [pa.Table, list]:
        """private method to retrieve data constant"""
        module = HandlerFactory.get_module(module_name=f"ds_capability.sample.{reference}")
        if reference.startswith("lookup_"):
            return AbstractSample._select_list(selection=module.data, size=size, seed=seed, shuffle=shuffle)
        df = pd.DataFrame.from_dict(module.data)
        idx = df.index.to_list()
        selection = AbstractSample._select_list(selection=idx, size=size, seed=seed, shuffle=shuffle)
        return pa.Table.from_pandas(df.iloc[selection].reset_index(drop=True))

    @staticmethod
    def _get_dataset(filename: str, size: int=None, shuffle: bool=True, seed: int=None, header: bool=True) -> [pa.Table, pa.Array]:
        """private method to retrieve a dataset"""
        header = 'infer' if header else None
        _path = Path(AbstractSample._full_path(filename))
        df = pd.read_csv(_path, encoding='latin1', header=header)
        idx = df.index.to_list()
        df = df.iloc[AbstractSample._select_list(selection=idx, size=size, seed=seed, shuffle=shuffle)]
        return pa.Table.from_pandas(df)

    @staticmethod
    def _select_list(selection: list, size: int=None, shuffle: bool=True, seed: int=None) -> pa.Array:
        """private method to select from a series
        :param shuffle:
        """
        size = size if isinstance(size, int) else len(selection)
        seed = seed if isinstance(seed, int) else int(time.time() * np.random.random())
        np.random.seed(seed)
        selection *= int(((size - 1) / len(selection)) + 1)
        if shuffle:
            idx = np.random.randint(0, len(selection), size=size)
            return pd.Series(selection).iloc[idx].values.tolist()[:size]
        return selection[:size]


class MappedSample(AbstractSample):

    def __dir__(self):
        rtn_list = []
        for m in dir(MappedSample):
            if not m.startswith('_'):
                rtn_list.append(m)
        return rtn_list

    @staticmethod
    def us_healthcare_organisations(size: int=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        return AbstractSample._get_constant(reference='map_us_healthcare_organisations', size=size, seed=seed,
                                            shuffle=shuffle)

    @staticmethod
    def us_healthcare_practitioner(size: int=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        seed = int(time.time() * np.random.random()) if not isinstance(seed, int) else seed
        generator = np.random.default_rng(seed=seed)
        a_sample = MappedSample.us_full_address(size=size, shuffle=False, seed=seed)
        size = a_sample.shape[0]
        p_sample = MappedSample.us_persona(female_bias=0.4, size=size, seed=seed)
        level = generator.choice(['MD', 'DNP'], size=size, p=[0.8, 0.2])
        df = pd.DataFrame()
        df['name'] = [f"{a} {b} {c}" for (a, b, c) in zip(p_sample['first_name'], p_sample['family_name'], level)]
        df['pcp_tax_id'] = np.linspace(100000000, 900000000, num=df.shape[0], dtype=int, endpoint=False)
        df['pcp_tax_id'] += np.random.randint(100, 999, size=df.shape[0])
        df = pd.concat([df, a_sample], axis='columns')
        df['address'] = df['address'].str.title()
        df['city'] = df['city'].str.title()
        if isinstance(shuffle, bool) and shuffle:
            df.sample(frac=1, random_state=seed).reset_index(drop=True)
        return pa.Table.from_pandas(df)

    @staticmethod
    def companies_fortune1000(size: int=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        return AbstractSample._get_constant(reference='map_companies_fortune1000', size=size, seed=seed,
                                            shuffle=shuffle)

    @staticmethod
    def companies_inc5000(size: int=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        return AbstractSample._get_constant(reference='map_companies_inc5000', size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def us_profession_rank(size: int=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        return AbstractSample._get_constant(reference='map_us_profession_detail_rank', size=size, seed=seed,
                                            shuffle=shuffle)

    @staticmethod
    def uk_postcodes_primary(size: int=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        return AbstractSample._get_constant(reference='map_uk_postcodes_primary', size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def us_city_area_code(size: int=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        # return AbstractSample._get_constant(reference='map_us_city_area_code', size=size, seed=seed, shuffle=shuffle)
        return AbstractSample._get_dataset(filename='map_us_city_area_code.csv', size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def us_city_zipcodes_rank(size: int=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        return AbstractSample._get_constant(reference='map_us_city_zipcodes_rank', size=size, seed=seed,
                                            shuffle=shuffle)

    @staticmethod
    def us_zipcodes_detail(size: int=None, shuffle: bool=False, state_filter: list=None, inc_military: bool=None,
                           seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) by default is shuffle but if false then returned in state abbreviation order.
        :param state_filter: (optional) a list of state abbreviations to filter on
        :param inc_military: (optional) if military zip codes should be included. Default to True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        inc_military = inc_military if isinstance(inc_military, bool) else True
        sample = AbstractSample._get_constant(reference='map_us_city_zipcodes_rank', shuffle=False).to_pandas()
        # trim
        if not inc_military:
            sample = sample[~sample['military']]
        if isinstance(state_filter, list):
            sample = sample[sample['state_id'].isin(state_filter)]
        sample.drop(columns=['ranking', 'incorporated', 'density', 'military', 'population', 'lat', 'lng', 'city_ascii',
                             'timezone'], inplace=True)
        # explode
        sample = sample.explode(column='zipcodes', ignore_index=True)
        size = size if isinstance(size, int) else sample.shape[0]
        rtn_idx = sample.index.to_list() * int(((size - 1) / sample.shape[0]) + 1)
        df_rtn = sample.iloc[rtn_idx]
        df_rtn = df_rtn.rename(columns={'zipcodes': 'zipcode', 'state_name': 'state', 'county_name': 'county',
                                        'state_id': 'state_abbr'})
        if shuffle:
            df_rtn = df_rtn.sample(frac=1, random_state=seed).reset_index(drop=True)
        else:
            df_rtn.sort_values(by=['state_abbr', 'county'], inplace=True)
        return pa.Table.from_pandas(df_rtn.iloc[:size])

    @staticmethod
    def us_phone_code(size: int=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        return AbstractSample._get_constant(reference='map_us_phone_code', size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def us_surname_rank(size: int=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        return AbstractSample._get_constant(reference='map_us_surname_rank', size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def us_forename_unisex(size: int=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        return AbstractSample._get_constant(reference='map_us_forename_unisex', size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def us_full_address(size: int=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns the first 'size' dataframe

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        return AbstractSample._get_constant(reference='map_us_full_address', size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def us_persona(size: int=None, female_bias: float=None, shuffle: bool=False, seed: int=None) -> pa.Table:
        """returns a frame of personas of 'size' where the female_bias is between zero and 1

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param female_bias: a female bias between 0 and 1 where 0 is zero females and 1 is all females
        :param shuffle: (optional) if not shuffled then returns in family_name alphabetical order
        :param seed: (optional) a seed value
        :return: the mapping DataFrame
        """
        sample = AbstractSample._get_constant(reference='map_us_forename_mf', shuffle=False)
        size = size if isinstance(size, int) else sample.shape[0]
        # generate the random index of the names
        generator = np.random.default_rng(seed=seed)
        # set male/female
        sample = sample.to_pandas()
        female_idx = sample[sample['Gender'] == 'F'].dropna().index.to_list()
        male_idx = sample[sample['Gender'] == 'M'].dropna().index.to_list()
        female_bias = female_bias if isinstance(female_bias, float) and 0 <= female_bias <= 1 else np.round(
            len(female_idx) / sample.shape[0], 5)
        female_bias += np.round(generator.uniform(low=-0.001, high=0.001), 4)
        female_size = int(np.round(female_bias * size, 0))
        male_size = size - female_size
        female_idx *= int(((female_size - 1) / len(female_idx)) + 1)
        female_second_idx = female_idx
        generator.shuffle(female_second_idx)
        male_idx *= int(((male_size - 1) / len(male_idx)) + 1)
        male_second_idx = male_idx
        generator.shuffle(male_second_idx)
        forename_idx = female_idx[:female_size] + male_idx[:male_size]
        generator.shuffle(forename_idx)
        middle_idx = female_second_idx[:female_size] + male_second_idx[:male_size]
        df_rtn = pd.DataFrame(columns=['first_name', 'middle_name', 'gender'])
        df_rtn['first_name'] = sample['Name'].iloc[forename_idx].values
        df_rtn['middle_name'] = sample['Name'].iloc[middle_idx].values
        df_rtn['gender'] = sample['Gender'].iloc[forename_idx].values
        df_rtn['middle_name'].iloc[0:int(df_rtn.shape[0] * generator.uniform(low=0.2, high=0.3))] = ''
        # set the surname
        df_rtn['family_name'] = Sample.us_surnames(size=size, shuffle=True, seed=seed)
        # set the email name
        df_rtn['email'] = pd.Series(
            [f"{a.lower()}.{b.lower()}" for (a, b) in zip(df_rtn['first_name'], df_rtn['family_name'])])
        # replace duplicates
        dup_email = df_rtn[df_rtn['email'].duplicated()].index.to_list()
        number = generator.integers(low=10, high=10000, size=len(dup_email))
        df_rtn['email'].iloc[dup_email] = pd.Series(
            [f"{a.lower()}{b}" for (a, b) in zip(df_rtn['family_name'], number)], dtype='string')
        domain = Sample.global_mail_domains(size=size, shuffle=True, seed=seed)
        df_rtn['email'] = pd.Series([f"{a}@{b}" for (a, b) in zip(df_rtn['email'], domain)])
        if shuffle:
            df_rtn = df_rtn.sample(frac=1, random_state=seed).reset_index(drop=True)
        else:
            df_rtn.sort_values(by=['family_name', 'first_name'], inplace=True)
        return pa.Table.from_pandas(df_rtn)


class Sample(AbstractSample):

    def __dir__(self):
        rtn_list = []
        for m in dir(Sample):
            if not m.startswith('_'):
                rtn_list.append(m)
        return rtn_list

    @staticmethod
    def us_surnames(size: int = None, shuffle: bool=True, seed: int = None) -> list:
        """returns a randomly selected list of surnames weighted on popularity over 150,000 of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """

        def divider(_size):
            weight_map = {1000000: 50000, 500000: 100000, 150000: 500000}
            for k, v in weight_map.items():
                if _size > k:
                    return v
            return 2000000

        size = size if isinstance(size, int) else 150000
        df = AbstractSample._get_constant(reference='map_us_surname_rank', seed=seed, shuffle=False).to_pandas()
        df['weight'] = [int(round((x / divider(size)), 0)) for x in df['count']]
        df_clean = df.where(df.weight > 1).dropna()
        result = [[x] * df_clean.weight.astype(int).iloc[x] for x in df_clean.index]
        idx = [j for i in result for j in i] + df.index.to_list()
        idx = AbstractSample._select_list(selection=idx, size=size, shuffle=shuffle, seed=seed)
        return df['name'].iloc[idx].to_list()

    @staticmethod
    def us_professions(size: int = None, shuffle: bool=True, seed: int = None) -> list:
        """returns a randomly selected list of profession names of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """
        df = AbstractSample._get_constant(reference='map_us_profession_detail_rank', seed=seed, shuffle=False).to_pandas()
        df.columns = ['occupation', 'total']
        size = size if isinstance(size, int) else df.shape[0]
        result = [[x] * df['total'].astype(int).iloc[x] for x in df.index]
        idx = [j for i in result for j in i]
        idx = AbstractSample._select_list(selection=idx, size=size, shuffle=shuffle, seed=seed)
        return df['occupation'].iloc[idx].to_list()

    @staticmethod
    def uk_street_types(size: int = None, shuffle: bool = True, seed: int = None) -> list:
        """returns a randomly selected list of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """
        selection = ['Road', 'Street', 'Way', 'Avenue', 'Drive', 'Lane', 'Grove', 'Gardens', 'Place', 'Circus',
                     'Crescent', 'Bypass', 'Close', 'Square', 'Hill', 'Mews', 'Vale', 'Rise', 'Row', 'Mead', 'Wharf']
        return Sample._select_list(selection=selection, size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def uk_cities(size: int = None, shuffle: bool=True, seed: int = None) -> list:
        """returns a randomly selected list of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """
        return AbstractSample._get_constant(reference='lookup_uk_city', size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def uk_postcode_district(size: int = None, shuffle: bool=True, seed: int = None) -> list:
        """returns a randomly selected list of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """
        return AbstractSample._get_constant(reference='lookup_uk_postcode_district', size=size, seed=seed,
                                            shuffle=shuffle)

    @staticmethod
    def us_street_names(size: int = None, shuffle: bool=True, seed: int = None) -> list:
        """returns a randomly selected list of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """
        return AbstractSample._get_constant(reference='lookup_us_street_names', size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def us_street_types(size: int = None, shuffle: bool = True, seed: int = None) -> list:
        """returns a randomly selected list of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """
        return AbstractSample._get_constant(reference='lookup_us_street_suffix', size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def us_cities(size: int = None, shuffle: bool=True, seed: int = None) -> list:
        """returns a randomly selected list of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """
        return AbstractSample._get_constant(reference='lookup_us_city', size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def us_zipcodes(size: int=None, shuffle: bool=True, seed: int = None) -> list:
        """returns a randomly selected list of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """
        return MappedSample.us_zipcodes_detail(size=size, seed=seed, shuffle=shuffle).column('zipcode').to_pylist()

    @staticmethod
    def us_states(size: int = None, shuffle: bool=True, seed: int = None) -> list:
        """returns a randomly selected list of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """
        selection = [
             'AA', 'AE', 'AK', 'AL', 'AP', 'AR', 'AS', 'AZ', 'CA', 'CO', 'CT', 'DC', 'DE', 'FL', 'FM', 'GA', 'GU',
             'HI', 'IA', 'ID', 'IL', 'IN', 'KS', 'KY', 'LA', 'MA', 'MD', 'ME', 'MH', 'MI', 'MN', 'MO', 'MP', 'MS',
             'MT', 'NC', 'ND', 'NE', 'NH', 'NJ', 'NM', 'NV', 'NY', 'OH', 'OK', 'OR', 'PA', 'PR', 'PW', 'RI', 'SC',
             'SD', 'TN', 'TX', 'UT', 'VA', 'VI', 'VT', 'WA', 'WI', 'WV', 'WY']
        return Sample._select_list(selection=selection, size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def global_mail_domains(size: int = None, shuffle: bool=True, seed: int = None) -> list:
        """returns a randomly selected list of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """
        selection = [
             "hotmail.com", "google.com", "facebook.com", "gmail.com", "googlemail.com", "msn.com", "verizon.net",
             "yahoo.com", "aol.com", "att.net", "comcast.net", "gmx.com", "mac.com", "me.com", "mail.com", "live.com",
             "sbcglobal.net", "email.com", "fastmail.fm", "games.com", "gmx.net", "hush.com", "hushmail.com",
             "icloud.com", "iname.com", "inbox.com", "lavabit.com", "love.com" "outlook.com", "pobox.com",
             "protonmail.com", "rocketmail.com", "safe-mail.net", "wow.com", "ygm.com", "ymail.com", "zoho.com",
             "yandex.com", "bellsouth.net", "charter.net", "cox.net", "earthlink.net", "juno.com"]
        return Sample._select_list(selection=selection, size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def british_mail_domains(size: int = None, shuffle: bool=True, seed: int = None) -> list:
        """returns a randomly selected list of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """
        selection = [
             "btinternet.com", "virginmedia.com", "blueyonder.co.uk", "freeserve.co.uk", "live.co.uk", "ntlworld.com",
             "o2.co.uk", "orange.net", "sky.com", "talktalk.co.uk", "tiscali.co.uk", "virgin.net", "wanadoo.co.uk",
             "bt.com", "yahoo.co.uk", "hotmail.co.uk"]
        return Sample._select_list(selection=selection, size=size, seed=seed, shuffle=shuffle)

    @staticmethod
    def complaint(size: int = None, shuffle: bool = True, seed: int = None) -> list:
        """returns a randomly selected list of size

        :param size: (optional) the size of the sample. If None then all the names are returned
        :param shuffle: (optional) if the list should be shuffled. Default is True
        :param seed: (optional) a seed value
        :return: a list of names
        """
        return AbstractSample._get_constant(reference='lookup_complaints', size=size, seed=seed, shuffle=shuffle)



from ds_capability.components.controller import Controller
import os
import pyarrow as pa


def run_repo_pipeline(canonical: pa.Table, repo_path: str, run_book: [str, list, dict]=None,
                      repeat: int=None, sleep: int=None, run_time: int=None, source_check_uri: str=None,
                      run_cycle_report: str=None, **kwargs):
    """"""
    for key, value in kwargs.items():
        key = key.upper()
        if key.startswith('HADRON_'):
            os.environ[key] = value

    controller = Controller.from_env(uri_pm_repo=repo_path, has_contract=False)

    controller.set_source_uri('event://internal_repo_pipeline/')
    controller.save_canonical(connector_name=Controller.CONNECTOR_SOURCE, canonical=canonical)
    controller.set_persist_uri('event://internal_repo_pipeline/')

    controller.run_controller(run_book=run_book, repeat=repeat, sleep=sleep, run_time=run_time,
                              source_check_uri=source_check_uri, run_cycle_report=run_cycle_report)
    # clean out environs
    for key in kwargs.keys():
        key = key.upper()
        if key in os.environ:
            del os.environ[key]

    return controller.load_canonical(connector_name=Controller.CONNECTOR_PERSIST, drop=True)




import ast
import unittest
import os
from pathlib import Path
import shutil
from datetime import datetime
from pprint import pprint

import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq

from ds_core.handlers.event_handlers import EventManager
from ds_core.properties.property_manager import PropertyManager
from ds_capability import FeatureEngineer, Controller, FeatureSelect
from ds_capability.components.pipeline_scripts import run_repo_pipeline

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class FeatureBuilderTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_controller(self):
        # setup
        os.environ['HADRON_CLEAN_SOURCE_URI'] = 'https://raw.githubusercontent.com/mwaskom/seaborn-data/master/titanic.csv'
        os.environ['HADRON_CLEAN_PERSIST_URI'] = 'event://demo/'

        fs = FeatureSelect.from_env('auto_clean', has_contract=False)
        fs.set_source_uri('${HADRON_CLEAN_SOURCE_URI}')
        fs.set_persist_uri('${HADRON_CLEAN_PERSIST_URI}')

        tbl = fs.load_source_canonical()
        tbl = fs.tools.auto_clean_header(tbl)
        tbl = fs.tools.auto_drop_noise(tbl)
        tbl = fs.tools.auto_drop_correlated(tbl)
        tbl = fs.tools.auto_drop_duplicates(tbl)
        tbl = fs.tools.auto_cast_types(tbl, include_category=False, include_bool=False)

        ctr = Controller.from_env(has_contract=False)
        ctr.register.feature_select('auto_clean')

        # reuse
        os.environ['HADRON_CLEAN_SOURCE_URI'] = 'https://raw.githubusercontent.com/mwaskom/seaborn-data/master/tips.csv'
        os.environ['HADRON_CLEAN_PERSIST_URI'] = 'event://demo/'
        os.environ['HADRON_PM_REPO'] = './working/contracts/'

        ctr = Controller.from_env()
        ctr.run_controller()

        # review
        ctr.set_persist_uri('event://demo/')
        print(ctr.table_report(ctr.load_persist_canonical()).to_string())

    def test_controller_scripts(self):
        repo_path = "https://github.com/project-hadron/hadron-asset-bank/blob/master/contracts/pyarrow/data_profiling"
        tbl = FeatureEngineer.from_memory().tools.get_synthetic_data_types(100)
        # pass the environ variables
        result = run_repo_pipeline(tbl, repo_path=repo_path,
                                   hadron_profiing_source_uri='s3://project-hadron-cs-repo/domain/synthetic/source/synthetic_type_10_000.parquet',
                                   hadron_profiing_persist_uri='s3://project-hadron-cs-repo/domain/synthetic/persist/synthetic_type_10_000_processed.parquet',
                                   hadron_data_profile_uri='s3://project-hadron-cs-repo/domain/synthetic/persist/quality.parquet',
                                   hadron_data_dictionary_uri='s3://project-hadron-cs-repo/domain/synthetic/persist/dictionary.parquet',
                                   hadron_data_schema_uri='s3://project-hadron-cs-repo/domain/synthetic/persist/schema.parquet')
        print(result.shape)

    def test_raise(self):
        startTime = datetime.now()
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))
        print(f"Duration - {str(datetime.now() - startTime)}")


def get_table(size: int=10, inc_null: bool=None):
    return FeatureEngineer.from_memory().tools.get_synthetic_data_types(size=size, extend=inc_null)

def set_service():
    EventManager().set('sample', get_table())
    fb1 = FeatureEngineer.from_env('task1', has_contract=False)
    # fb1.set_source_uri('./working/source/hadron_synth_origin.pq')
    fb1.set_persist()
    tbl = fb1.tools.get_synthetic_data_types(size=10)
    _ = fb1.tools.get_noise(size=10, num_columns=2, canonical=tbl)
    fb1.run_component_pipeline()
    pprint(pm_view('feature_build', 'task1'))
    fb2 = FeatureEngineer.from_env('task2', has_contract=False)
    fb2.set_source_uri(fb1.get_persist_contract().raw_uri)
    fb2.set_persist()
    source_tbl = fb2.load_source_canonical()
    _ = fb2.tools.correlate_number(canonical=source_tbl, header='num', column_name='corr')
    fb2.run_component_pipeline()
    controller = Controller.from_env(has_contract=False)
    controller.intent_model.feature_build(canonical=None, task_name='task1', intent_level='task1_tr')
    controller.intent_model.wrangle(canonical=None, task_name='task2', intent_level='task2_wr')

def pm_view(capability: str, task: str, section: str=None):
    uri = os.path.join(os.environ['HADRON_PM_PATH'], f"hadron_pm_{capability}_{task}.parquet")
    tbl = pq.read_table(uri)
    tbl = tbl.column(0).combine_chunks()
    result = ast.literal_eval(tbl.to_pylist()[0]).get(capability,{}).get(task,{})
    return result.get(section, {}) if isinstance(section, str) and section in result.keys() else result



if __name__ == '__main__':
    unittest.main()



import collections
import pandas as pd
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
from ds_capability.components.commons import Commons


# noinspection PyArgumentList
class DataDiscovery(object):

    @staticmethod
    def outliers_iqr(values: pa.Array, k_factor: float=None):
        """ The interquartile range (IQR), also called the midspread, middle 50%, or H‑spread, is a measure of
        statistical dispersion, being equal to the difference between 75th and 25th percentiles, or between upper
        and lower quartiles

        The IQR can be used to identify outliers by defining limits on the sample values that are a factor k of the
        IQR below the 25th percentile or above the 75th percentile. The common value for the factor k is the value 1.5.
        A factor k of 3 or more can be used to identify values that are extreme outliers.

        :param values: the values to be assessed
        :param k_factor: the k factor to apply beyond the 25th and 75th percentile, defaults to 1.5
        :return: a tuple of the lower and upper outliers as a list of index of the values
        """
        k_factor = k_factor if isinstance(k_factor, float) and 0 < k_factor <= 10 else 1.5
        values = values.drop_null()
        # calculate interquartile range
        q25, q75 = pc.quantile(values, [0.25, 0.75])
        iqr = pc.subtract(q75, q25)
        # calculate the outlier cutoff
        cut_off = pc.multiply(iqr, k_factor)
        lower, upper = pc.subtract(q25, cut_off), pc.add(q75, cut_off)
        # identify outliers
        return pc.or_(pc.greater(values, upper), pc.less(values, lower))

    @staticmethod
    def outliers_empirical(values: pa.Array, std_width: int=None):
        """ The empirical rule states that for a normal distribution, nearly all of the data will fall within three
        standard deviations of the mean.

        Given mu and sigma, a simple way to identify outliers is to compute a z-score for every value, which is
        defined as the number of standard deviations away value is from the mean. The 68–95–99.7 rule, also known as
        the empirical rule, is a shorthand used to remember the percentage of values that lie within a band around
        the mean in a normal distribution with a width of two, four and six standard deviations, respectively

        :param values: the values to be assessed
        :param std_width: (optional) A standard deviation away from the mean. Default is 3
        :return: a list of index of outliers
        """
        values = values.drop_null()
        std_width = std_width if isinstance(std_width, int) and 2 <= std_width <= 6 else 3
        data_mean, data_std = pc.mean(values), pc.stddev(values)
        # identify outliers
        cut_off = pc.multiply(data_std, std_width)
        lower, upper = pc.subtract(data_mean, cut_off), pc.add(data_mean, cut_off)
        # identify outliers
        return pc.or_(pc.greater(values, upper), pc.less(values, lower))

    @staticmethod
    def data_quality(canonical: pa.Table, nulls_threshold: float=None, dom_threshold: float=None,
                     cat_threshold: int=None, discrete_threshold: int=None, capped_at: int=None,
                     stylise: bool=None):
        """ Analyses a dataset, passed as a DataFrame and returns a quality summary

        :param canonical: The dataset, as a DataFrame.
        :param cat_threshold: The threshold for the max number of unique categories. Default is 20
        :param discrete_threshold: The threshold for the max number of unique numeric. Default is 20
        :param dom_threshold: The threshold limit of a dominant value. Default 0.98
        :param nulls_threshold: The threshold limit of a nulls value. Default 0.9
        :param capped_at: the row and column cap or 0 to ignore. default 5_000_000
        :param stylise: if the output is stylised for jupyter display
        :return: pd.DataFrame
        """
        # defaults
        cat_threshold = cat_threshold if isinstance(cat_threshold, int) else 20
        discrete_threshold = discrete_threshold if isinstance(discrete_threshold, int) else 20
        dom_threshold = dom_threshold if isinstance(dom_threshold, float) and 0 <= dom_threshold <= 1 else 0.98
        nulls_threshold = nulls_threshold if isinstance(nulls_threshold, float) and 0 <= nulls_threshold <= 1 else 0.95
        stylise = stylise if isinstance(stylise, bool) else False
        cap = capped_at if isinstance(capped_at, int) else 5_000_000
        if canonical.num_rows*canonical.num_columns > cap > 0:
            row_count = int(round(cap / canonical.num_columns, 0))
            canonical = canonical.slice(0, row_count)
        # cast the values
        tbl = Commons.table_cast(canonical, cat_max=cat_threshold)
        # dictionary
        _null_columns = []
        _dom_columns = []
        _sparce_columns = []
        _date_columns = []
        _bool_columns = []
        _cat_columns = []
        _disc_columns = []
        _num_columns = []
        _int_columns = []
        _str_columns = []
        _nest_columns = []
        _other_columns = []
        _key_columns = []
        for n in tbl.column_names:
            c = tbl.column(n).combine_chunks()
            if pa.types.is_nested(c.type):
                _nest_columns.append(n)
                continue
            elif pa.types.is_string(c.type) and c.null_count != tbl.num_rows:
                _str_columns.append(n)
            elif pa.types.is_integer(c.type) and c.null_count != tbl.num_rows:
                _int_columns.append(n)
            elif pa.types.is_floating(c.type) and c.null_count != tbl.num_rows:
                _num_columns.append(n)
            elif pa.types.is_boolean(c.type) and c.null_count != tbl.num_rows:
                _bool_columns.append(n)
            elif pa.types.is_timestamp(c.type) or pa.types.is_time(c.type) and c.null_count != tbl.num_rows:
                _date_columns.append(n)
            elif pa.types.is_dictionary(c.type):
                _cat_columns.append(n)
                c = c.dictionary_decode()
            else:
                _other_columns.append(n)
                # nulls and
            if c.null_count/tbl.num_rows > nulls_threshold:
                _null_columns.append(n)
            elif c.null_count / tbl.num_rows > 0.66:
                _sparce_columns.append(n)
            elif pc.count_distinct(c.drop_null()).as_py() == pc.count(c).as_py():
                _key_columns.append(n)
            elif 1-(pc.count_distinct(c.drop_null()).as_py()/pc.count(c).as_py()) > dom_threshold:
                _dom_columns.append(n)
        # dictionary
        _usable_columns = _date_columns + _bool_columns + _cat_columns + _num_columns + _int_columns + _str_columns
        _null_avg = len(_null_columns) / canonical.num_columns
        _dom_avg = len(_dom_columns) / canonical.num_columns
        _quality_avg = int(round(100 - (((_null_avg + _dom_avg) / 2) * 100), 0))
        _usable = int(round((len(_usable_columns) / canonical.num_columns) * 100, 2))
        # duplicate
        _dup_columns = []
        for i in range(0, len(tbl.column_names)):
            col_1 = tbl.column_names[i]
            for col_2 in tbl.column_names[i + 1:]:
                if tbl.column(col_1).equals(tbl.column(col_2)):
                    _dup_columns.append(col_2)
        # time
        _dt_today = pd.to_datetime('today')
        mem_usage = canonical.get_total_buffer_size()
        _tbl_mem = f"{mem_usage >> 20} Mb" if mem_usage >> 20 > 0 else f"{mem_usage} bytes"
        mem_usage = pa.total_allocated_bytes()
        _tot_mem = f"{mem_usage >> 20} Mb" if mem_usage >> 20 > 0 else f"{mem_usage} bytes"
        report = {
            'timestamp': {'readable': _dt_today.strftime('%d %B %Y %I:%M %p'),
                          'semantic': _dt_today.strftime('%Y-%m-%dT%H:%M:%SZ')},
            'score': {'quality_avg': f"{_quality_avg}%", 'usability_avg': f"{_usable}%"},
            'data_shape': {'tbl_memory': _tbl_mem, 'total_allocated': _tot_mem,
                           'rows': canonical.num_rows, 'columns': canonical.num_columns},
            'data_type': {'floating': len(_num_columns), 'integer': len(_int_columns),
                          'category': len(_cat_columns), 'datetime': len(_date_columns),
                          'bool': len(_bool_columns),'string': len(_str_columns),
                          'nested': len(_nest_columns), 'others': len(_other_columns)},
            'usability': {'mostly_null': len(_null_columns),
                          'predominance': len(_dom_columns),
                          'sparse': len(_sparce_columns),
                          'duplicate': len(_dup_columns),
                          'candidate_keys': len(_key_columns)}
        }
        # convert to multi-index DataFrame
        result = pd.DataFrame.from_dict(report, orient="index").stack().to_frame()
        result = pd.DataFrame(result[0].values.tolist(), index=result.index, columns=['summary'])
        result['summary'] = result['summary'].apply(str).str.replace('.0', '', regex=False)
        result = result.reset_index(names=['sections', 'elements'])
        if stylise:
            return Commons.report(result, index_header='sections', bold=['elements'])
        reference = [_num_columns, _int_columns, _cat_columns, _date_columns, _bool_columns, _str_columns, _nest_columns,
                 _other_columns, _null_columns, _dom_columns, _sparce_columns, _dup_columns, _key_columns]
        return Commons.table_append(pa.Table.from_pandas(result),
                                    pa.table([pa.array([[]] * 8 + reference, pa.list_(pa.string()))], names=['columns']))


    @staticmethod
    def data_dictionary(canonical: pa.Table, table_cast: bool=None, display_width: int=None, stylise: bool=None,
                        capped_at: int=None, ordered: bool=None):
        """ The data dictionary for a given canonical

        :param canonical: The canonical to interpret
        :param table_cast: attempt to cast columns to the content
        :param display_width: the width of the display
        :param stylise: if the output is stylised for jupyter display
        :param capped_at: the row and column cap or 0 to ignore. default 5_000_000
        :param ordered: if the columns are sorted by name.
        :return: a pa.Table or stylised pandas
        """
        display_width = display_width if isinstance(display_width, int) else 50
        stylise = stylise if isinstance(stylise, bool) else False
        cap = capped_at if isinstance(capped_at, int) else 5_000_000
        if canonical.num_rows*canonical.num_columns > cap > 0:
            row_count = int(round(cap / canonical.num_columns, 0))
            canonical = canonical.slice(0, row_count)
        record = []
        labels = [f'Attributes', 'DataType', 'Nulls', 'Dominate', 'Valid', 'Unique', 'Observations']
        # attempt cast
        if isinstance(table_cast, bool) and table_cast:
            canonical = Commons.table_cast(canonical)
        column_names = canonical.column_names
        if isinstance(ordered, bool) and ordered:
            column_names.sort()
        for c in column_names:
            column = canonical.column(c).combine_chunks()
            if pa.types.is_nested(column.type):
                s = str(column.slice(0,20).to_pylist())
                if len(s) > display_width:
                    s = s[:display_width] + "..."
                record.append([c,'nested',0,0,1,1,s])
                continue
            # data type
            line = [c,
                    'category' if pc.starts_with(str(column.type), 'dict').as_py() else str(column.type),
                    # null percentage
                    round(column.null_count / canonical.num_rows, 3)]
            # dominant percentage
            arr_vc = column.value_counts()
            value = arr_vc.filter(pc.equal(arr_vc.field(1), pc.max(arr_vc.field(1)))).field(1)[0].as_py()
            line.append(round(value / canonical.num_rows, 3))
            # valid
            line.append(pc.sum(column.is_valid()).as_py())
            # unique
            line.append(pc.count(column.unique()).as_py())
            # observations
            vc = column.drop_null().value_counts()
            if pa.types.is_dictionary(column.type):
                t = pa.table([vc.field(1), vc.field(0).dictionary], names=['v', 'n']).sort_by([("v", "descending")])
            else:
                t = pa.table([vc.field(1), vc.field(0)], names=['v', 'n']).sort_by([("v", "descending")])
            s = str(t.column('n').to_pylist())
            if len(s) > display_width:
                s = s[:display_width] + "..."
            line.append(s)
            record.append(line)
        df = pd.DataFrame(record, columns=labels)
        if stylise:
            style = [{'selector': 'th', 'props': [('font-size', "120%"), ("text-align", "center")]},
                     {'selector': '.row_heading, .blank', 'props': [('display', 'none;')]}]
            df_style = df.style.set_table_styles(style)
            _ = df_style.map(DataDiscovery._highlight_null_dom, subset=['Nulls', 'Dominate'])
            _ = df_style.map(lambda x: 'color: white' if x > 0.98 else 'color: black', subset=['Nulls', 'Dominate'])
            _ = df_style.map(DataDiscovery._dtype_color, subset=['DataType'])
            _ = df_style.map(DataDiscovery._color_unique, subset=['Unique'])
            _ = df_style.map(lambda x: 'color: white' if x < 2 else 'color: black', subset=['Unique'])
            _ = df_style.format({'Nulls': "{:.1%}", 'Dominate': '{:.1%}'})
            _ = df_style.set_caption(f"dataset has {canonical.num_columns} columns")
            _ = df_style.set_properties(subset=['Attributes'],  **{'font-weight': 'bold', 'font-size': "120%"})
            return df_style
        return pa.Table.from_pandas(df)

    @staticmethod
    def data_schema(canonical: pa.Table, table_cast: bool=None, capped_at: int=None, stylise: bool=None):
        """ The data dictionary for a given canonical

        :param canonical: The canonical to interpret
        :param table_cast: (optional) attempt to cast columns to the content
        :param capped_at: the row and column cap or 0 to ignore. default 5_000_000
        :param stylise: (optional) if the output is stylised for jupyter display
        :return: a pa.Table or stylised pandas
        """
        stylise = stylise if isinstance(stylise, bool) else False
        cap = capped_at if isinstance(capped_at, int) else 5_000_000
        if canonical.num_rows*canonical.num_columns > cap > 0:
            row_count = int(round(cap / canonical.num_columns, 0))
            canonical = canonical.slice(0, row_count)
        if isinstance(table_cast, bool) and table_cast:
            canonical = Commons.table_cast(canonical)
        record = []
        for n in canonical.column_names:
            c = canonical.column(n).combine_chunks()
            if pa.types.is_nested(c.type) or pa.types.is_binary(c.type) or pc.equal(c.null_count, canonical.num_rows).as_py():
                continue
            if pa.types.is_dictionary(c.type):
                vc = c.drop_null().value_counts()
                t = pa.table([vc.field(1), vc.field(0).dictionary], names=['v','n']).sort_by([("v", "descending")])
                record.append([n, 'categories', t.column('n').to_pylist()])
                _ = pc.round(pc.divide_checked(t.column('v').cast(pa.float64()), pc.sum(t.column('v'))),3).to_pylist()
                record.append([n, 'frequency', _])
                record.append([n, 'type', 'category'])
                record.append([n, 'measure', 'discrete'])
                record.append([n, 'nulls', c.null_count])
                record.append([n, 'valid', pc.sum(c.is_valid()).as_py()])
                record.append([n, 'null_proportions', (c.null_count/len(c))])
                record.append([n, 'valid_proportions', ((pc.sum(c.is_valid()).as_py())/len(c))])
                unique_count =  sum(1 for _, count in collections.Counter(c.to_pylist()).items() if count == 1)
                record.append([n, 'unique', unique_count])
                record.append([n, 'unique_proportions', (unique_count/len(c))])
                distinct_count = len(c.unique())
                record.append([n, 'distinct', distinct_count])
                record.append([n, 'distinct_proportions', (distinct_count/len(c))])
            elif pa.types.is_integer(c.type) or pa.types.is_floating(c.type):
                precision = Commons.column_precision(c)
                intervals = DataDiscovery.to_discrete_intervals(column=c, granularity=5, categories=['A','B','C','D','E'])
                vc = intervals.dictionary_encode().drop_null().value_counts()
                t = pa.table([vc.field(1), vc.field(0).dictionary], names=['v','n']).sort_by([("n", "ascending")])
                record.append([n, 'intervals', ['lower','low','mid','high','higher']])
                _ = pc.round(pc.divide_checked(t.column('v').cast(pa.float64()), pc.sum(t.column('v'))),3).to_pylist()
                record.append([n, 'frequency', _])
                record.append([n, 'type', c.type])
                record.append([n, 'measure', 'temporal' if pa.types.is_temporal(c.type) else 'continuous'])
                record.append([n, 'nulls', c.null_count])
                record.append([n, 'valid', pc.sum(c.is_valid()).as_py()])
                record.append([n, 'null_proportions', (c.null_count/len(c))])
                record.append([n, 'valid_proportions', ((pc.sum(c.is_valid()).as_py())/len(c))])
                unique_count =  sum(1 for _, count in collections.Counter(c.to_pylist()).items() if count == 1)
                record.append([n, 'unique', unique_count])
                record.append([n, 'unique_proportions', (unique_count/len(c))])
                distinct_count = len(c.unique())
                record.append([n, 'distinct', distinct_count])
                record.append([n, 'distinct_proportions', (distinct_count/len(c))])
                record.append([n, 'mean', pc.round(pc.mean(c),precision).as_py()])
                record.append([n, 'std', pc.round(pc.sqrt(pc.variance(c)),precision).as_py()])
                record.append([n, 'max', pc.round(pc.max(c),precision).as_py()])
                record.append([n, '75%', pc.round(pc.quantile(c,0.75),precision).to_pylist()[0]])
                record.append([n, '50%', pc.round(pc.quantile(c,0.5),precision).to_pylist()[0]])
                record.append([n, '25%', pc.round(pc.quantile(c,0.25),precision).to_pylist()[0]])
                record.append([n, 'min', pc.round(pc.min(c),precision).as_py()])
                # skew
                values = c.slice(0, 100000).to_pandas()
                record.append([n, 'skew', round(values.skew(), 3)])
                s_bins = [-np.inf, -1, -0.5, -0.1, 0.1, 0.5, 1, np.inf]
                s_names = ['highly left', 'medium left', 'light left', 'normal', 'light right', 'medium right',
                           'highly right']
                _skew_bin = list(pd.cut([values.skew()], s_bins, labels=s_names))[0]
                record.append([n, 'skew bias', _skew_bin])
                # kurtosis
                record.append([n, 'kurtosis', round(values.kurtosis(), 3)])
                k_bins = [-np.inf, 2.9, 3.1, np.inf]
                k_names = ['platykurtic', 'mesokurtic', 'leptokurtic']
                _kurt_bin = list(pd.cut([values.kurtosis()], k_bins, labels=k_names))[0]
                record.append([n, 'kurtosis_tail', _kurt_bin])
            elif pa.types.is_boolean(c.type):
                vc = c.drop_null().value_counts()
                t = pa.table([vc.field(1), vc.field(0)], names=['v', 'n']).sort_by([("n", "ascending")])
                record.append([n, 'boolean', t.column('n').to_pylist()])
                _ = pc.round(pc.divide_checked(t.column('v').cast(pa.float64()), pc.sum(t.column('v'))),3).to_pylist()
                record.append([n, 'frequency', _])
                record.append([n, 'type', c.type])
                record.append([n, 'measure', 'binary'])
                record.append([n, 'nulls', c.null_count])
                record.append([n, 'valid', pc.sum(c.is_valid()).as_py()])
                record.append([n, 'null_proportions', (c.null_count/len(c))])
                record.append([n, 'valid_proportions', ((pc.sum(c.is_valid()).as_py())/len(c))])
                unique_count =  sum(1 for _, count in collections.Counter(c.to_pylist()).items() if count == 1)
                record.append([n, 'unique', unique_count])
                record.append([n, 'unique_proportions', (unique_count/len(c))])
                distinct_count = len(c.unique())
                record.append([n, 'distinct', distinct_count])
                record.append([n, 'distinct_proportions', (distinct_count/len(c))])
            elif pa.types.is_timestamp(c.type) or pa.types.is_time(c.type) or pa.types.is_date(c.type):
                _ = pa.array(Commons.date2value(c.to_pylist()))
                intervals = DataDiscovery.to_discrete_intervals(column=_, granularity=5, categories=['A','B','C','D','E'])
                vc = intervals.dictionary_encode().drop_null().value_counts()
                t = pa.table([vc.field(1), vc.field(0).dictionary], names=['v','n']).sort_by([("n", "ascending")])
                record.append([n, 'intervals', ['older','old','mid','new','newer']])
                _ = pc.round(pc.divide_checked(t.column('v').cast(pa.float64()), pc.sum(t.column('v'))),3).to_pylist()
                record.append([n, 'frequency', _])
                record.append([n, 'type', c.type])
                record.append([n, 'measure', 'temporal' if pa.types.is_temporal(c.type) else 'continuous'])
                record.append([n, 'nulls', c.null_count])
                record.append([n, 'valid', pc.sum(c.is_valid()).as_py()])
                record.append([n, 'null_proportions', (c.null_count/len(c))])
                record.append([n, 'valid_proportions', ((pc.sum(c.is_valid()).as_py())/len(c))])
                unique_count =  sum(1 for _, count in collections.Counter(c.to_pylist()).items() if count == 1)
                record.append([n, 'unique', unique_count])
                record.append([n, 'unique_proportions', (unique_count/len(c))])
                distinct_count = len(c.unique())
                record.append([n, 'distinct', distinct_count])
                record.append([n, 'distinct_proportions', (distinct_count/len(c))])
                record.append([n, 'oldest', pc.min(c).as_py()])
                record.append([n, 'newest', pc.max(c).as_py()])
            elif pa.types.is_string(c.type):
                record.append([n, 'string', []])
                record.append([n, 'frequency', []])
                record.append([n, 'type', c.type])
                record.append([n, 'measure', ''])
                record.append([n, 'nulls', c.null_count])
                record.append([n, 'valid', pc.sum(c.is_valid()).as_py()])
                record.append([n, 'null_proportions', (c.null_count/len(c))])
                record.append([n, 'valid_proportions', ((pc.sum(c.is_valid()).as_py())/len(c))])
                unique_count =  sum(1 for _, count in collections.Counter(c.to_pylist()).items() if count == 1)
                record.append([n, 'unique', unique_count])
                record.append([n, 'unique_proportions', (unique_count/len(c))])
                distinct_count = len(c.unique())
                record.append([n, 'distinct', distinct_count])
                record.append([n, 'distinct_proportions', (distinct_count/len(c))])
        df = pd.DataFrame(record, columns=['attributes', 'elements', 'values'])
        df['values'] = df['values'].astype(str)
        if stylise:
            return Commons.report(df, index_header='attributes', bold=['elements'])
        return pa.Table.from_pandas(df)

    @staticmethod
    def to_discrete_intervals(column: pa.Array, granularity: [int, float, list]=None, lower: [int, float]=None,
                              upper: [int, float]=None, categories: list=None, precision: int=None) -> pa.Array:
        """ creates discrete intervals from continuous values """
        # intend code block on the canonical
        granularity = granularity if isinstance(granularity, (int, float, list)) or granularity == 0 else 5
        granularity = len(categories) if isinstance(categories, list) else granularity
        precision = precision if isinstance(precision, int) else 5
        lower = lower if isinstance(lower, (int, float)) else pc.min(column).as_py()
        # firstly get the granularity
        upper = upper if isinstance(upper, (int, float)) else pc.max(column).as_py()
        s_values = column.to_pandas()
        if lower >= upper:
            upper = lower
            granularity = [(lower, upper, 'both')]
        if isinstance(granularity, (int, float)):
            # if granularity float then convert frequency to intervals
            if isinstance(granularity, float):
                # make sure frequency goes beyond the upper
                _end = upper + granularity - (upper % granularity)
                periods = pd.interval_range(start=lower, end=_end, freq=granularity).drop_duplicates()
                periods = periods.to_tuples().to_list()
                granularity = []
                while len(periods) > 0:
                    period = periods.pop(0)
                    if len(periods) == 0:
                        granularity += [(period[0], period[1], 'both')]
                    else:
                        granularity += [(period[0], period[1], 'left')]
            # if granularity int then convert periods to intervals
            else:
                periods = pd.interval_range(start=lower, end=upper, periods=granularity).drop_duplicates()
                granularity = periods.to_tuples().to_list()
        if isinstance(granularity, list):
            if all(isinstance(value, tuple) for value in granularity):
                if len(granularity[0]) == 2:
                    granularity[0] = (granularity[0][0], granularity[0][1], 'both')
                granularity = [(t[0], t[1], 'right') if len(t) == 2 else t for t in granularity]
            elif all(isinstance(value, float) and 0 < value < 1 for value in granularity):
                quantiles = list(set(granularity + [0, 1.0]))
                boundaries = s_values.quantile(quantiles).values
                boundaries.sort()
                granularity = [(boundaries[0], boundaries[1], 'both')]
                granularity += [(boundaries[i - 1], boundaries[i], 'right') for i in range(2, boundaries.size)]
            else:
                granularity = (lower, upper, 'both')
        granularity = [(np.round(p[0], precision), np.round(p[1], precision), p[2]) for p in granularity]
        # now create the categories
        conditions = []
        for interval in granularity:
            lower, upper, closed = interval
            if str.lower(closed) == 'neither':
                conditions.append((s_values > lower) & (s_values < upper))
            elif str.lower(closed) == 'right':
                conditions.append((s_values > lower) & (s_values <= upper))
            elif str.lower(closed) == 'both':
                conditions.append((s_values >= lower) & (s_values <= upper))
            else:
                conditions.append((s_values >= lower) & (s_values < upper))
        if isinstance(categories, list) and len(categories) == len(conditions):
            choices = categories
        else:
            if s_values.dtype.name.startswith('int'):
                choices = [f"{int(i[0])}->{int(i[1])}" for i in granularity]
            else:
                choices = [f"{i[0]}->{i[1]}" for i in granularity]
        # noinspection PyTypeChecker
        rtn_list = np.select(conditions, choices, default=None).tolist()
        return pa.StringArray.from_pandas(rtn_list)


    @staticmethod
    def _dtype_color(dtype: str):
        """Apply color to types"""
        if str(dtype).startswith('cat'):
            color = '#208a0f'
        elif str(dtype).startswith('int'):
            color = '#0f398a'
        elif str(dtype).startswith('float'):
            color = '#2f0f8a'
        elif str(dtype).startswith('date'):
            color = '#790f8a'
        elif str(dtype).startswith('bool'):
            color = '#08488e'
        elif str(dtype).startswith('str'):
            color = '#761d38'
        else:
            return ''
        return 'color: %s' % color

    @staticmethod
    def _highlight_null_dom(x: str):
        x = float(x)
        if not isinstance(x, float) or x < 0.65:
            return ''
        elif x < 0.85:
            color = '#ffede5'
        elif x < 0.90:
            color = '#fdcdb9'
        elif x < 0.95:
            color = '#fcb499'
        elif x < 0.98:
            color = '#fc9576'
        elif x < 0.99:
            color = '#fb7858'
        elif x < 0.997:
            color = '#f7593f'
        else:
            color = '#ec382b'
        return 'background-color: %s' % color

    @staticmethod
    def _color_unique(x: str):
        x = int(x)
        if not isinstance(x, int):
            return ''
        elif x < 2:
            color = '#ec382b'
        elif x < 3:
            color = '#a1cbe2'
        elif x < 5:
            color = '#84cc83'
        elif x < 10:
            color = '#a4da9e'
        elif x < 20:
            color = '#c1e6ba'
        elif x < 50:
            color = '#e5f5e0'
        elif x < 100:
            color = '#f0f9ed'
        else:
            return ''
        return 'background-color: %s' % color



from ds_core.handlers.abstract_handlers import AbstractSourceHandler, ConnectorContract
from ds_core.handlers.abstract_handlers import HandlerFactory, AbstractPersistHandler
import pandas as pd
import pyarrow as pa

__author__ = 'Darryl and Sekhar'


class MysqlSourceHandler(AbstractSourceHandler):
    """ This handler class uses both SQLAlchemy and pymysql. Together, SQLAlchemy and pymysql provide a powerful
    toolset for working with databases and faster connectivity. SQLAlchemy allows developers to interact with MySQL
    using Python code, while MySQL provides the database functionality needed to store and retrieve data efficiently.

        URI example
            uri = "mysql://name:password@host:port/database?param2=param1&param2=param2"

        params:
            query: (optional) a source query added as a parameter or kward
            table: (optional) the sql table assuming full dataframe persist. By default 'hadron_default' is used
            if_exists: (optional) How to behave if the table already exists.‘fail’, ‘replace’, ‘append’. Default replace

    """

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Handler passing the Connector Contract """
        # required module import
        self.pymysql = HandlerFactory.get_module('pymysql')
        self.sqlalchemy = HandlerFactory.get_module('sqlalchemy')
        self.sqlalchemy_schema = HandlerFactory.get_module('sqlalchemy.schema')
        super().__init__(connector_contract)
        # reset to use dialect
        _kwargs = {**self.connector_contract.kwargs, **self.connector_contract.query}
        self._sql_query = _kwargs.pop('query', "")
        self._sql_table = _kwargs.pop('table', "hadron_default")
        self._if_exists = _kwargs.pop('if_exists', 'replace')
        # add dialect
        address = self.connector_contract.address.replace('mysql:', 'mysql+pymysql:')
        self._engine = self.sqlalchemy.create_engine(address, **_kwargs)
        self._update_time = pd.Timestamp("1970-01-01")
        self._changed_flag = True

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['mysql']

    def exists(self) -> bool:
        """If the table exists"""
        try:
            meta = self.sqlalchemy_schema.MetaData()
            meta.reflect(bind=self._engine)
            if self._sql_table in meta.tables.keys():
                return True
            return False
        except self.pymysql.Error:
            return False

    def has_changed(self) -> bool:
        """ if the table has changed. Only works with certain implementations"""
        # TODO: Add in change logic here
        database = str(self.connector_contract.path[1:])
        table = str(self._sql_table)
        query = f"SELECT UPDATE_TIME " \
                f"FROM information_schema.tables " \
                f"WHERE TABLE_SCHEMA = '{database}' AND TABLE_NAME = '{table}';"
        with self._engine.connect() as con:
            result = con.execute(query)
        last_update = pd.to_datetime((next(result))[0])
        if last_update is None or self._update_time < last_update:
            self._changed_flag = True
        else:
            self._changed_flag = False
        return self._changed_flag

    def reset_changed(self, changed: bool = False):
        """ manual reset to say the table has been seen. This is automatically called if the file is loaded"""
        changed = changed if isinstance(changed, bool) else False
        self._changed_flag = changed

    def load_canonical(self, **kwargs) -> dict:
        """ returns the canonical dataset based on the Connector Contract """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Connector Contract is not valid")
        try:
            connect = self._engine.connect()
            query = self._sql_query if len(self._sql_query) > 0 else f"SELECT * FROM {self._sql_table}"
            rtn_tbl = pa.Table.from_pandas(pd.read_sql(query, con=connect, **kwargs))
            connect.close()
            return rtn_tbl
        except self.pymysql.Error as error:
            raise ConnectionError(f"Failed to load the canonical to MySQL because {error}")


class MysqlPersistHandler(MysqlSourceHandler, AbstractPersistHandler):
    # a MySQL persist handler

    def persist_canonical(self, canonical: pa.Table, **kwargs) -> bool:
        """ persists the canonical dataset"""
        return self.backup_canonical(canonical=canonical, table=self._sql_table, **kwargs)

    def backup_canonical(self, canonical: pa.Table, table: str, **kwargs) -> bool:
        """ creates a backup of the canonical to an alternative table  """
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        try:
            _if_exists = self._if_exists
            _params = kwargs
            _if_exists = _params.pop('if_exists', self._if_exists)
            connect = self._engine.connect()
            canonical = canonical.to_pandas()
            canonical.to_sql(con=connect, name=table, if_exists=_if_exists, index=False, **_params)
            connect.close()
            return True
        except self.pymysql.Error as error:
            raise ConnectionError(f"Failed to save the canonical to MySQL because {error}")

    def remove_canonical(self) -> bool:
        """removes the table and content"""
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _cc = self.connector_contract
        try:
            query = f"DROP TABLE IF EXISTS {self._sql_table};"
            with self._engine.connect() as con:
                con.execute(query)
        except self.pymysql.Error:
            return False
        return True



import unittest
import os
from pathlib import Path
import shutil
from datetime import datetime
import pandas as pd
import pyarrow as pa

import pyarrow.compute as pc
from ds_capability.components.commons import Commons

from ds_capability.handlers.pyarrow_handlers import PyarrowSourceHandler, PyarrowPersistHandler
from ds_core.handlers.abstract_handlers import ConnectorContract
from ds_core.properties.property_manager import PropertyManager

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class FeatureBuilderTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_parquet(self):
        tbl = get_table()
        uri = os.path.join(os.environ['HADRON_DEFAULT_PATH'], 'test.parquet')
        cc = ConnectorContract(uri, 'module_name', 'handler')
        handler = PyarrowPersistHandler(cc)
        handler.persist_canonical(tbl)
        result = handler.load_canonical()
        self.assertEqual(tbl.column_names, result.column_names)
        self.assertEqual(tbl.shape, result.shape)
        self.assertEqual(tbl.schema, result.schema)

    def test_csv(self):
        tbl = get_table()
        uri = os.path.join(os.environ['HADRON_DEFAULT_PATH'], 'test.csv')
        cc = ConnectorContract(uri, 'module_name', 'handler')
        handler = PyarrowPersistHandler(cc)
        handler.persist_canonical(tbl)
        result = handler.load_canonical()
        self.assertEqual(tbl.column_names, result.column_names)
        self.assertEqual(tbl.shape, result.shape)

    def test_csv_https(self):
        tbl = get_table()
        uri = "https://raw.githubusercontent.com/mwaskom/seaborn-data/master/titanic.csv"
        cc = ConnectorContract(uri, 'module_name', 'handler')
        handler = PyarrowPersistHandler(cc)
        result = handler.load_canonical()
        self.assertEqual((891, 15), result.shape)

    def test_json(self):
        tbl = get_table()
        uri = os.path.join(os.environ['HADRON_DEFAULT_PATH'], 'test.json')
        cc = ConnectorContract(uri, 'module_name', 'handler')
        handler = PyarrowPersistHandler(cc)
        # tbl
        handler.persist_canonical(tbl)
        result = handler.load_canonical()
        print(result.shape)
        # properties
        cfg = {'feature_build': {'data_profiling': {'description': '', 'version': '0.0.1', 'status': 'discovery', 'connectors': {'pm_feature_build_data_profiling': {'raw_uri': 'working/contracts/hadron_pm_feature_build_data_profiling.json', 'raw_module_name': 'ds_capability.handlers.pyarrow_handlers', 'raw_handler': 'PyarrowPersistHandler', 'raw_version': '0.0.1', 'raw_kwargs': {}, 'aligned': False}, 'template_source': {'raw_uri': 'working/data', 'raw_module_name': 'ds_capability.handlers.pyarrow_handlers', 'raw_handler': 'PyarrowSourceHandler', 'raw_version': '0.0.1', 'raw_kwargs': {}, 'aligned': False}, 'template_persist': {'raw_uri': 'working/data', 'raw_module_name': 'ds_capability.handlers.pyarrow_handlers', 'raw_handler': 'PyarrowPersistHandler', 'raw_version': '0.0.1', 'raw_kwargs': {}, 'aligned': False}, 'primary_source': {'raw_uri': '${HADRON_PROFILING_SOURCE_URI}', 'raw_module_name': 'ds_capability.handlers.pyarrow_handlers', 'raw_handler': 'PyarrowPersistHandler', 'raw_version': '0.0.1', 'raw_kwargs': {}, 'aligned': False}, 'primary_persist': {'raw_uri': 'event://profiling', 'raw_module_name': 'ds_core.handlers.event_handlers', 'raw_handler': 'EventPersistHandler', 'raw_version': '0.0.1', 'raw_kwargs': {}, 'aligned': False}, 'quality': {'raw_uri': '${HADRON_DATA_QUALITY_URI}', 'raw_module_name': 'ds_capability.handlers.pyarrow_handlers', 'raw_handler': 'PyarrowPersistHandler', 'raw_version': '0.0.1', 'raw_kwargs': {}, 'aligned': False}, 'dictionary': {'raw_uri': '${HADRON_DATA_DICTIONARY_URI}', 'raw_module_name': 'ds_capability.handlers.pyarrow_handlers', 'raw_handler': 'PyarrowPersistHandler', 'raw_version': '0.0.1', 'raw_kwargs': {}, 'aligned': False}, 'schema': {'raw_uri': '${HADRON_DATA_SCHEMA_URI}', 'raw_module_name': 'ds_capability.handlers.pyarrow_handlers', 'raw_handler': 'PyarrowPersistHandler', 'raw_version': '0.0.1', 'raw_kwargs': {}, 'aligned': False}}, 'intent': {'primary': {'0': {'build_profiling': {'profiling': 'quality', 'connector_name': 'quality', 'save_intent': False}}, '1': {'build_profiling': {'profiling': 'dictionary', 'connector_name': 'dictionary', 'save_intent': False}}, '2': {'build_profiling': {'profiling': 'schema', 'connector_name': 'schema', 'save_intent': False}}}}, 'snapshot': {}, 'run_book': {}, 'meta': {'module': ['ds_capability', 'managers', 'feature_build_property_manager'], 'class': 'FeatureBuildPropertyManager'}, 'knowledge': {'describe': {}, 'intent': {}, 'schema': {}}}}, 'config_meta': {'uid': 'c18ffafd-f28d-4835-81f8-bd5d580fb54a', 'create': '2023-10-14 13:33:59.006304', 'modify': '2023-10-14 13:33:59.151534', 'release': '0.8.8'}}
        tbl = pa.Table.from_pylist([cfg])
        tbl = Commons.table_flatten(tbl)
        handler.persist_canonical(tbl)
        result = handler.load_canonical()
        print(result.shape)
        # swagger
        cfg = {'columnProfile': [{'customMetricsProfile': [{'name': 'string', 'value': 0}], 'distinctCount': 0, 'distinctProportion': 0, 'duplicateCount': 0, 'firstQuartile': 0, 'histogram': {'boundaries': [{}], 'frequencies': [{}]}, 'interQuartileRange': 0, 'max': {}, 'maxLength': 0, 'mean': 0, 'median': 0, 'min': {}, 'minLength': 0, 'missingCount': 0, 'missingPercentage': 0, 'name': 'string', 'nonParametricSkew': 0, 'nullCount': 0, 'nullProportion': 0, 'stddev': 0, 'sum': 0, 'thirdQuartile': 0, 'timestamp': 0, 'uniqueCount': 0, 'uniqueProportion': 0, 'validCount': 0, 'valuesCount': 0, 'valuesPercentage': 0, 'variance': 0}], 'systemProfile': [{'operation': 'UPDATE', 'rowsAffected': 0, 'timestamp': 0}], 'tableProfile': {'columnCount': 0, 'createDateTime': '2019-08-24T14:15:22Z', 'profileSample': 0, 'profileSampleType': 'PERCENTAGE', 'rowCount': 0, 'sizeInByte': 0, 'timestamp': 0}}
        tbl = pa.Table.from_pylist([cfg])
        tbl = Commons.table_flatten(tbl)
        handler.persist_canonical(tbl)
        result = handler.load_canonical()
        print(result.shape)

    def test_txt(self):
        tbl = get_table()
        uri = os.path.join(os.environ['HADRON_DEFAULT_PATH'], 'test.txt')
        cc = ConnectorContract(uri, 'module_name', 'handler')
        handler = PyarrowPersistHandler(cc)
        handler.persist_canonical(tbl)
        result = handler.load_canonical()
        self.assertEqual(tbl.column_names, result.column_names)
        self.assertEqual(tbl.shape, result.shape)

    def test_raise(self):
        startTime = datetime.now()
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))
        print(f"Duration - {str(datetime.now() - startTime)}")



def get_table():
    num = pa.array([1.0, None, 5.0, -0.46421, 3.5, 7.233, -2], pa.float64())
    val = pa.array([1, 2, 3, 4, 5, 6, 7], pa.int64())
    date = pc.strptime(["2023-01-02 04:49:06", "2023-01-02 04:57:12", None, None, "2023-01-02 05:23:50", None, None],
                       format='%Y-%m-%d %H:%M:%S', unit='us')
    text = pa.array(["Blue", "Green", None, 'Red', 'Orange', 'Yellow', 'Pink'], pa.string())
    binary = pa.array([True, True, None, False, False, True, False], pa.bool_())
    cat = pa.array([None, 'M', 'F', 'M', 'F', 'M', 'M'], pa.string()).dictionary_encode()
    return pa.table([num, val, date, text, binary, cat], names=['num', 'int', 'date', 'text', 'bool', 'cat'])


if __name__ == '__main__':
    unittest.main()



import inspect
import random
import re
import pyarrow as pa
import pyarrow.compute as pc
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

from ds_capability.components.commons import Commons
from ds_capability.intent.common_intent import CommonsIntentModel
from ds_capability.intent.abstract_feature_select_intent import AbstractFeatureSelectIntentModel


class FeatureSelectIntent(AbstractFeatureSelectIntentModel, CommonsIntentModel):

    """This class represents feature selection intent actions focusing on dimensionality and
    specifically columnar reduction. Its purpose is to disregard irrelevant features to remove,
    amongst other things, constants, duplicates and statistically uninteresting columns.

    As an early stage data pipeline process, FeatureSelect focuses on data preprocessing, and as
    such is a filter step for extracting features of interest.
    """

    def auto_clean_header(self, canonical: pa.Table, case: str=None, rename_map: [dict, list, str]=None,
                          replace_spaces: str=None, save_intent: bool=None, intent_level: [int, str]=None,
                          intent_order: int=None, replace_intent: bool=None, remove_duplicates: bool=None) -> pa.Table:
        """ clean the headers of a Table replacing space with underscore. This also allows remapping and case selection

        :param canonical: the pa.Table
        :param rename_map: (optional) a dict of name value pairs, a fixed length list of column names or connector name
        :param case: (optional) changes the headers to lower, upper, title. if none of these then no change
        :param replace_spaces: (optional) character to replace spaces with. Default is '_' (underscore)
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: pa.Table.
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # Code block for intent
        # auto mapping
        if isinstance(rename_map, str):
            if self._pm.has_connector(rename_map):
                handler = self._pm.get_connector_handler(rename_map)
                mapper = handler.load_canonical()
                if mapper.shape[1] == 1:
                    rename_map = mapper.iloc[:, 0].values.tolist()
                else:
                    rename_map = dict(zip(mapper.iloc[:, 0].values, mapper.iloc[:, 1].values))
            else:
                mapper=None
        # map the headers
        if isinstance(rename_map, dict):
            names = [rename_map.get(item,item)  for item in canonical.column_names]
            canonical = canonical.rename_columns(names)
        if isinstance(rename_map, list) and len(rename_map) == canonical.num_columns:
            tbl = canonical.rename_columns(rename_map)
        # tidy
        headers = []
        for s in canonical.column_names:
            s = re.sub(r"[^\w\s]", '', s)
            s = re.sub(r"\s+", '_', s)
            headers.append(s)
        # convert case
        if isinstance(case, str):
            headers = pa.array(headers)
            if case.lower() == 'lower':
                headers = pc.ascii_lower(headers)
            elif case.lower() == 'upper':
                headers = pc.ascii_upper(headers)
            elif case.lower() == 'title':
                headers = pc.ascii_title(headers)
            headers = headers.to_pylist()
        # return table with new headers
        return canonical.rename_columns(headers)

    def auto_cast_types(self, canonical: pa.Table, include_category: bool=None, category_max: int=None, include_bool: bool=None,
                        include_timestamp: bool=None, tm_format: str=None, tm_units: str=None, tm_tz: str=None, save_intent: bool=None,
                        intent_level: [int, str]=None, intent_order: int=None, replace_intent: bool=None,
                        remove_duplicates: bool=None) -> pa.Table:
        """ attempts to cast the columns of a table to its appropriate type. Categories boolean and timestamps
        are toggled on and off with the include parameters being true or false.

        :param canonical: the pa.Table
        :param include_category: (optional) if categories should be cast.  Default True
        :param category_max: (optional) the max number of unique values to consider categorical
        :param include_bool: (optional) if booleans should be cast. Default True
        :param include_timestamp: (optional) if categories should be cast.  Default True
        :param tm_format: (optional) if not standard, the format of the dates, example '%m-%d-%Y %H:%M:%S'
        :param tm_units: (optional) units to cast timestamp. Options are 's', 'ms', 'us', 'ns'
        :param tm_tz: (optional) timezone to cast timestamp
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: pa.Table.
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # Code block for intent
        return Commons.table_cast(canonical, inc_cat=include_category, cat_max=category_max, inc_bool=include_bool,
                                  inc_time=include_timestamp, dt_format=tm_format, units=tm_units, tz=tm_tz)

    def auto_reinstate_nulls(self, canonical: pa.Table, nulls_list=None, headers: [str, list]=None, drop: bool=None,
                             data_type: [str, list]=None, regex: [str, list]=None, save_intent: bool=None,
                             intent_level: [int, str]=None, intent_order: int=None, replace_intent: bool=None,
                             remove_duplicates: bool=None) -> pa.Table:
        """ automatically reinstates nulls in a string that have been masked with alternate values such as space
        or question-mark. By default, the nulls list is ['',' ','NaN','nan','None','null','Null','NULL']

        :param canonical: the pa.Table
        :param nulls_list: (optional) potential null values to replace with a null.
        :param headers: a list of headers to drop or filter on type
        :param drop: to drop or not drop the headers
        :param data_type: the column types to include or exclude. Default None else int, float, bool, object, 'number'
        :param regex: a regular expression to search the headers
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: pa.Table.
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # Code block for intent
        nulls_list = nulls_list if isinstance(nulls_list, list) else ['', ' ', 'NaN', 'nan', 'None', 'null', 'Null',
                                                                      'NULL']

        selected_headers = Commons.filter_headers(canonical, headers=headers, d_types=data_type, regex=regex, drop=drop)
        rtn_tbl = None
        for n in selected_headers:
            c = canonical.column(n).combine_chunks()
            if pa.types.is_string(c.type):
                mask = pc.is_in(c, pa.array(nulls_list))
                c = pc.if_else(mask, None, c)
                canonical = Commons.table_append(canonical, pa.table([c], names=[n]))
        return canonical

    def auto_drop_noise(self, canonical: pa.Table, nulls_threshold: float=None, nulls_list: [bool, list]=None,
                        drop_predominant: bool=None, drop_empty_row: bool=None, drop_unknown: bool=None,
                        save_intent: bool=None, intent_level: [int, str]=None, intent_order: int=None,
                        replace_intent: bool=None, remove_duplicates: bool=None) -> pa.Table:
        """ auto removes columns that are at least 0.998 percent np.NaN, a single value, std equal zero or have a
        predominant value greater than the default 0.998 percent.

        :param canonical: the pa.Table
        :param nulls_threshold:  (optional) The threshold limit of a nulls value. Default 0.95
        :param nulls_list:  (optional) can be boolean or a list:
                    if boolean and True then null_list equals ['NaN', 'nan', 'null', '', 'None', ' ']
                    if list then this is considered potential null values.
        :param drop_predominant:  (optional) drop columns that have a predominant value of the given predominant max
        :param drop_empty_row:  (optional) also drop any rows where all the values are empty
        :param drop_unknown:  (optional) drop objects that are not string types such as binary
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: pa.Table.
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # Code block for intent
        nulls_threshold = nulls_threshold if isinstance(nulls_threshold, float) and 0 <= nulls_threshold <= 1 else 0.95
        drop_unknown = drop_unknown if isinstance(drop_unknown, bool) else False
        # drop knowns
        to_drop = []
        for n in canonical.column_names:
            c = canonical.column(n).combine_chunks()
            if pa.types.is_dictionary(c.type):
                c = c.dictionary_decode()
            if pa.types.is_nested(c.type) or pa.types.is_list(c.type) or pa.types.is_struct(c.type):
                to_drop.append(n)
            elif c.null_count / canonical.num_rows > nulls_threshold:
                to_drop.append(n)
            elif pc.count(pc.unique(c)).as_py() <= 1:
                to_drop.append(n)
        return canonical.drop_columns(to_drop)

    def auto_sample_rows(self, canonical: pa.Table, size: int, save_intent: bool=None, intent_level: [int, str]=None,
                         intent_order: int=None, replace_intent: bool=None, remove_duplicates: bool=None) -> pa.Table:
        """ auto samples rows of a canonical returning a randomly selected subset of the canonical based on size.

        :param canonical: the pa.Table
        :param size: the randomly selected subset size of the canonical
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: pa.Table.
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # Code block for intent
        if canonical.num_rows > size > 0:
            sample = random.sample(range(canonical.num_rows), k=int(size))
            canonical = canonical.take(sample)
        return canonical

    def auto_drop_columns(self, canonical: pa.Table, headers: [str, list]=None, d_types: [str, list]=None,
                          regex: [str, list]=None, drop: bool=None, save_intent: bool=None,
                          intent_level: [int, str]=None, intent_order: int=None, replace_intent: bool=None,
                          remove_duplicates: bool=None) -> pa.Table:
        """ auto removes columns that are selected.

        :param canonical: the pa.Table
        :param headers: (optional) a filter of headers from the 'other' dataset
        :param drop: (optional) to drop or not drop the headers if specified
        :param d_types: (optional) a filter on data type for the 'other' dataset. int, float, bool, object
        :param regex: (optional) a regular expression to search the headers. example '^((?!_amt).)*$)' excludes '_amt'
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: pa.Table.
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # Code block for intent
        to_drop = Commons.filter_headers(canonical, headers=headers, regex=regex, d_types=d_types, drop=drop)
        return canonical.drop_columns(to_drop)

    def auto_drop_duplicates(self, canonical: pa.Table, save_intent: bool=None, intent_level: [int, str]=None,
                             intent_order: int=None, replace_intent: bool=None,
                             remove_duplicates: bool=None) -> pa.Table:
        """ Removes columns that are duplicates of each other

        :param canonical: the pa.Table
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: Canonical,.
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # Code block for intent
        to_drop = []
        for i in range(0, len(canonical.column_names)):
            col_1 = canonical.column_names[i]
            for col_2 in canonical.column_names[i + 1:]:
                if canonical.column(col_1).equals(canonical.column(col_2)):
                    to_drop.append(col_2)
        return canonical.drop_columns(to_drop)

    def auto_drop_correlated(self, canonical: pa.Table, threshold: float=None, save_intent: bool=None,
                             intent_level: [int, str]=None, intent_order: int=None, replace_intent: bool=None,
                             remove_duplicates: bool=None) -> pa.Table:
        """ uses 'brute force' techniques to remove highly correlated numeric columns based on the threshold,
        set by default to 0.95.

        :param canonical: the pa.Table
        :param threshold: (optional) threshold correlation between columns. default 0.95
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: Canonical,.
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # Code block for intent
        threshold = threshold if isinstance(threshold, float) and 0 < threshold < 1 else 0.95
        # extract numeric columns
        tbl_filter = Commons.filter_columns(canonical, d_types=[pa.int64(), pa.int32(), pa.float64(), pa.float32()])
        df_filter = tbl_filter.to_pandas()
        to_drop = set()
        corr_matrix = df_filter.corr()
        for i in range(len(corr_matrix.columns)):
            for j in range(i):
                if abs(corr_matrix.iloc[i, j]) > threshold:  # we are interested in absolute coeff value
                    col_name = corr_matrix.columns[i]  # getting the name of column
                    to_drop.add(col_name)
        return canonical.drop_columns(to_drop)

    def auto_aggregate(self, canonical: pa.Table, action: str, headers: [str, list]=None, d_types: [str, list]=None,
                       regex: [str, list]=None, drop: bool=None, to_header: str=None, drop_aggregated: bool=None,
                       precision: int=None, save_intent: bool=None, intent_level: [int, str]=None,
                       intent_order: int=None, replace_intent: bool=None, remove_duplicates: bool=None) -> pa.Table:
        """ given a set of columns, aggregates those columns based upon the aggregation action given.
        The actions are 'sum', 'prod', 'count', 'min', 'max', 'mean', 'list', 'list_first', 'list_last'.
        'list_first' and 'list_last' return the first or last value in a list.

        :param canonical: the pa.Table
        :param action: an aggregation action such as count or list_first.
        :param headers: (optional) a filter of headers from the 'other' dataset
        :param drop: (optional) to drop or not drop the headers if specified
        :param d_types: (optional) a filter on data type for the 'other' dataset. int, float, bool, object
        :param regex: (optional) a regular expression to search the headers. example '^((?!_amt).)*$)' excludes '_amt'
        :param to_header: (optional) an optional name to call the column
        :param drop_aggregated: (optional) drop the aggregation headers
        :param precision: the value precision of the return values
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: Canonical,.
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # Code block for intent
        drop_aggregated = drop_aggregated if isinstance(drop_aggregated, bool) else False
        tbl = Commons.filter_columns(canonical, headers=headers, d_types=d_types, regex=regex, drop=drop)
        headers = tbl.column_names
        df = tbl.to_pandas()
        if action not in ['sum', 'prod', 'count', 'min', 'max', 'mean', 'list', 'list_first', 'list_last']:
            raise ValueError("The only values are 'sum','prod','count','min','max','mean','list','list_first','list_last'")
        # Code block for intent
        precision = precision if isinstance(precision, int) else 3
        if action.startswith('list'):
            rtn_values = df.loc[:, headers].values.tolist()
            rtn_values = [[x for x in y if x is not None] for y in rtn_values]
            if action.endswith('_first'):
                rtn_values = [x[0] if len(x) > 0 else None for x in rtn_values]
            if action.endswith('_last'):
                rtn_values = [x[-1] if len(x) > 0 else None for x in rtn_values]
        else:
            rtn_values = eval(f"df.loc[:, headers].{action}(axis=1)", globals(), locals()).round(precision).to_list()
        to_header = to_header if isinstance(to_header, str) else next(self.label_gen)
        if drop_aggregated:
            canonical = canonical.drop_columns(headers)
        return Commons.table_append(canonical, pa.table([pa.array(rtn_values)], names=[to_header]))

    def auto_projection(self, canonical: pa.Table, headers: list=None, drop: bool=None, n_components: [int, float]=None,
                        seed: int=None, save_intent: bool=None, intent_level: [int, str]=None, intent_order: int=None,
                        replace_intent: bool=None, remove_duplicates: bool=None, **kwargs) -> pa.Table:
        """Principal component analysis (PCA) is a linear dimensionality reduction using Singular Value Decomposition
        of the data to project it to a lower dimensional space.

        :param canonical: the pa.Table
        :param headers: (optional) a list of headers to select (default) or drop from the dataset
        :param drop: (optional) if True then srop the headers. False by default
        :param n_components: (optional) Number of components to keep.
        :param seed: (optional) placeholder
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :param kwargs: additional parameters to pass the PCA model
        :return: a pd.DataFrame
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # Code block for intent
        headers = Commons.list_formatter(headers)
        sample = Commons.filter_columns(canonical, headers=headers, drop=drop, d_types=[pa.float64(), pa.int64()])
        sample = self.auto_drop_noise(sample, nulls_threshold=0.3)
        sample = Commons.table_fill_null(sample)
        if not sample or len(sample) == 0:
            return canonical
        n_components = n_components if isinstance(n_components, (int, float)) \
                                       and 0 < n_components < sample.shape[1]  else sample.shape[1]
        sample = sample.to_pandas(split_blocks=True)
        # standardise
        scaler = StandardScaler()
        train = scaler.fit_transform(sample)
        # pca
        pca = PCA(n_components=n_components, **kwargs)
        train = pca.fit_transform(train)
        gen = Commons.label_gen(prefix='pca_')
        names = []
        for n in range(train.shape[1]):
            names.append(next(gen))
        tbl = pa.Table.from_arrays(train.T, names=names)
        canonical = canonical.drop_columns(sample.columns)
        return Commons.table_append(canonical, tbl)

    def auto_append_tables(self, canonical: pa.Table, other: pa.Table=None, headers: [str, list]=None,
                           data_types: [str, list]=None, regex: [str, list]=None, drop: bool=None,
                           other_headers: [str, list]=None, other_data_type: [str, list]=None,
                           other_regex: [str, list]=None, other_drop: bool=None, seed: int=None,
                           save_intent: bool=None, intent_level: [int, str]=None, intent_order: int=None,
                           replace_intent: bool=None, remove_duplicates: bool=None) -> pa.Table:
        """ Appends the canonical table with other

        :param canonical: a pa.Table
        :param other: (optional) the pa.Table or connector to join. This is the dominant table and will replace like named columns
        :param headers: (optional) headers to select
        :param data_types: (optional) data types to select. use PyArrow data types eg 'pa.string()'
        :param regex: (optional) a regular expression
        :param drop: (optional) if True then drop the headers. False by default
        :param other_headers: other headers to select
        :param other_data_type: other data types to select. use PyArrow data types eg 'pa.string()'
        :param other_regex: other regular expression
        :param other_drop: if True then drop the other headers
        :param seed: (optional) placeholder
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # Code block for intent
        canonical = self._get_canonical(canonical)
        other = self._get_canonical(other)
        canonical = Commons.filter_columns(canonical, headers=headers, d_types=data_types, regex=regex, drop=drop)
        if other is None:
            return canonical
        other = Commons.filter_columns(other, headers=other_headers, d_types=other_data_type, regex=other_regex,
                                       drop=other_drop)
        if canonical.num_rows > other.num_rows:
            df = other.to_pandas()
            df = df.sample(n=canonical.num_rows, random_state=seed, ignore_index=True, replace=True)
            other = pa.Table.from_pandas(df)
        else:
            other = other.slice(0, canonical.num_rows)
        # append
        return Commons.table_append(canonical, other)



import unittest
import os
from pathlib import Path
import shutil
import ast
from datetime import datetime

import duckdb
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq
from ds_core.properties.property_manager import PropertyManager
from ds_capability import *
from ds_capability.components.commons import Commons
from ds_capability.handlers.duckdb_handlers import DuckdbSourceHandler, DuckdbPersistHandler

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class DuckdbTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'parquet'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_default(self):
        fe = FeatureEngineer.from_memory()
        tbl = fe.tools.get_synthetic_data_types(10, seed=1)
        source_uri = f"duckdb://"
        persist_uri = f"duckdb://"
        fe.set_source_uri(source_uri)
        # persist
        fe.set_persist_uri(persist_uri)
        fe.save_persist_canonical(tbl)
        # source
        result = fe.load_source_canonical()
        self.assertEqual((10,6), result.shape)
        # distinct
        distinct_sql = "SELECT num, int FROM @"
        distinct_uri = f"duckdb://?sql_query={distinct_sql}"
        fe.add_connector_uri('distinct', distinct_uri)
        result = fe.load_canonical('distinct')
        self.assertEqual((10,2), result.shape)

    def test_table(self):
        fe = FeatureEngineer.from_memory()
        tbl = fe.tools.get_synthetic_data_types(10, seed=1)
        source_uri = f"duckdb://?table=my_test"
        persist_uri = f"duckdb://?table=my_test"
        fe.set_source_uri(source_uri)
        # persist
        fe.set_persist_uri(persist_uri)
        fe.save_persist_canonical(tbl)
        # source
        result = fe.load_source_canonical()
        self.assertEqual((10, 6), result.shape)

    def test_s3(self):
        fe = FeatureEngineer.from_memory()
        tbl = fe.tools.get_synthetic_data_types(100)
        s3_uri = 's3://project-hadron-cs-repo/domain/synthetic/source/duckdb_tbl_100.parquet'
        persist_uri = f"duckdb://{s3_uri}"
        # persist_uri = f"duckdb://?table={s3_uri}&sql={persist_sql}"
        fe.set_persist_uri(persist_uri)
        fe.save_persist_canonical(tbl)
        s3_uri = 's3://project-hadron-cs-repo/domain/synthetic/source/synthetic_sample.parquet'
        source_uri = f"duckdb:///?table={s3_uri}"
        fe.set_source_uri(source_uri)
        result = fe.load_source_canonical()
        print(result.shape)

    def test_s3_direct(self):
        fe = FeatureEngineer.from_memory()
        tbl = fe.tools.get_synthetic_data_types(100)
        s3_uri = 's3://project-hadron-cs-repo/domain/synthetic/source/duckdb_tbl_100.parquet'
        conn = duckdb.connect(':default:')
        conn.execute(f"""
            INSTALL httpfs;
            LOAD httpfs;
            SET s3_region='{os.environ.get('AWS_DEFAULT_REGION')}';
            SET s3_access_key_id='{os.environ.get('AWS_ACCESS_KEY_ID')}';
            SET s3_secret_access_key='{os.environ.get('AWS_SECRET_ACCESS_KEY')}';
        """)
        conn.execute(f"COPY tbl TO '{s3_uri}'")
        # export
        result = conn.execute(f"SELECT * FROM read_parquet('{s3_uri}')").arrow()
        print(result.shape)

    def test_local_direct(self):
        fe = FeatureEngineer.from_memory()
        tbl = fe.tools.get_synthetic_data_types(100)
        conn = duckdb.connect(':default:')
        conn.execute("CREATE OR REPLACE TABLE my_arrow AS SELECT * FROM tbl;")
        conn.execute("CREATE OR REPLACE TABLE dup AS SELECT * FROM tbl;")
        result = conn.execute("SELECT * FROM my_arrow;").arrow()
        print(result.shape)
        result = conn.execute("CALL duckdb_tables()").arrow()
        print(result.column('table_name'))
        print(pc.is_in('my_arrow', result.column('table_name')).as_py())


    def test_raise(self):
        startTime = datetime.now()
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))
        print(f"Duration - {str(datetime.now() - startTime)}")


def tprint(t: pa.table, headers: [str, list] = None, d_type: [str, list] = None, regex: [str, list] = None):
    _ = Commons.filter_columns(t.slice(0, 10), headers=headers, d_types=d_type, regex=regex)
    print(Commons.table_report(_).to_string())


if __name__ == '__main__':
    unittest.main()
